/**
 * Implementation of class "GeneticAlgorithm".
 *
 * Author: Stefan Buettcher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 **/

package ga;

import examples.*;
import helperclasses.*;
import ttclasses.*;
import java.io.*;
import java.text.*;
import java.util.*;

/*****
 * The <tt>GeneticAlgorithm</tt> class is the main class used for looking
 * for an optimal (or sub-optimal) solution of the timetabling problem.
 *****/
public final class GeneticAlgorithm implements TTSolver {

  /** This is the standard Genetic Algorithm, using generational reproduction,
      i.e. the whole population (or a part of it, in case of using the ELITE
      operator) is replaced by the offspring. **/
  public static final int GENERATIONAL = 1;
  /** The second run mode is steady-state where the weakest individual is
      replaced in every step. **/
  public static final int STEADY_STATE = 2;
  /** The third run mode is a Hybrid Genetic Algorithm. Parallel to the actual
      GA-based evolution process, we try some hill-climbing methods in order
      to improve the performance of the individuals inside the current population.
      The hybrid mode is based on the steady-state GA. **/
  public static final int HYBRID = 3;

  /** Canonical selection means selection with probability proportional to
      the individual's fitness value. Some offset and scaling may,
      however, be applied to the raw fitness values. **/
  public static final int CANONICAL_SELECTION = 1;
  /** Rank-based selection selects each individual with probability
      proportional to <tt>n + 1 - rank</tt>, where <tt>n</tt> is the
      size of the current population, while <tt>rank</tt> is the
      individual's ranking (considering the fitness values) within
      the current population. **/
  public static final int RANK_BASED_SELECTION = 2;
  /** Tournament selection selects the best individual within a randomly-
      chosen subset of the current population. The size of the subset
      can be specified by calling <tt>setTournamentSize(int)</tt>. **/
  public static final int TOURNAMENT_SELECTION = 3;


  // If we use fitness-proportional selection, fitness values have to be scaled in some way.

  /** Goldberg scaling means the method proposed by Goldberg in "GAs in S, O, M/L", pp. 77-79. **/
  public static final int GOLDBERG_SCALING = 1;
  /** This scaling method defines the fitness value of the weakest individual 1, the fitness
      value of the fittest individual (1 + sqrt(n)) with linear interpolation in between. **/
  public static final int SQRT_SCALING = 2;
  /** This scaling method defines the fitness value of the weakest individual 1, the fitness
      value of the fittest individual 2 with linear interpolation in between. **/
  public static final int DEFAULT_SCALING = 2;

  private int runMode;                 // generational GA or steady-state
  private int startMode;               // the run mode that the GA has been started with
  private int populationSize;          // the number of individuals (time tables)
                                       // per generation
  private int selectionMethod;         // which selection method shall be used?
  private int scalingMethod;           // which type of scaling in case of fitness-proportional selection?
  private int tournamentSize;          // number of individuals involved in tournament selection
  private boolean alwaysKeepBest;      // always keep the best individual
  private boolean alwaysReplaceWorst;  // always replace the worst individual in steady-state mode?
                                       // canonical steady-state would be: only replace if new offspring is better
  private int crossOverIndividuals;    // how many individuals are necessary for
                                       // crossover? normal case: 2
  private int elitePercentage;         // relative size of the set of individuals that are
                                       // kept in the population by the ELITE operator
  private double mutationRate;         // the mutation rate for the individuals' gene sequence
  private double directedMutationRate; // the probability for directed mutation
  private int crossOverPoints;         // number of cross-over points (0 = uniform cross-over)

  private boolean autoParameters;      // set the parameter values automatically?

  private double repairRate;           // if an individual is to be repaired, then
                                       // repair that percentage of violations encountered

  private double climbRate;            // this is an indicator of how much the chromosomes shall
                                       // try to climb when the hill-climbing operator is used
                                       // with timetables, this is the ratio of events that the
                                       // the local-search operator shall be used on

  private boolean adjustSearchSpace;   // allows the GA to fix certain bits in the bitstring representation
                                       // so that these bits will not be changed by the mutation operator

  private double hybridProtocol[];     // in case of hybrid mode, all success and failure rates will be
                                       // resorded here

  private static final double hybridPeriod = 100;
  // weigh the outcome of every use of the hybrid operators with 1/hybridPeriod
  private double hybridCrossoverProbability, hybridClimbProbability, hybridRepairProbability;
  private double hybridCrossoverSuccessRate, hybridClimbSuccessRate, hybridRepairSuccessRate;
  // these variables will be used to determine use of hybrid operators

  private boolean islandMode;          // if set to true, indicates that this GA instance is part of a
                                       // distributed GA run
  private IslandFerry ferry;           // the ferry that transports individuals to other islands (machines)
  private int portNumber;              // which port number shall be used for listening?
  private int migrationInterval;       // specifies with what frequency individuals migrate between populations
                                       // migrationInterval == 2 means: every 2nd generation in the generational GA
                                       // and every (2*populationSize)th step in the steady-state GA
  private String harborFile;           // filename of the file which contains the harbor information
                                       // (i.e. list of all harbors involved in the evolution process)
  private byte[] immigrant;            // an individual received from another island

  // termination criteria
  private int maxGenerationCount;      // terminate execution as soon as this limit is exceeded
  private int timeLimit;               // over-all time limit in seconds
  private int stagnationLimit;         // terminate if for that many seconds there has been no improvement
  
  private int hybridActivationLimit;   // if for that many seconds there has been no improvement, RUN_MODE
                                       // will automatically switched to HYBRID
  private int hybridActivationTime;    // Regardless of stagnation, switch to HYBRID mode after that many seconds

  private PrintStream outputStream;    // this is where we write all the interesting stuff to

  private boolean localSearchAfterEvolution; // perform some local search at the end?

  private GeneticIndividual[] population;  // the current population
  private double[] fitnessValues;          // array of fitness values of the
                                           // members of the current population

  public  GeneticIndividual bestCreature = null;
          // the best individual at all. this very instance does not necessarily have to
          // be a member of the current population
  public  double bestFitness, previousBestFitness, bestSoFar;
          // fitness values of the best individual of the current and the
          // previous generation as well es the best one seen so far

  /** This is the sum of all individual's fitness values before scaling. **/
  public  double populationFitness;
  /** This is the sum of all individual's fitness values after scaling. **/
  public  double scaledFitness;

  private int generationCount;  // index number of the current population / generation

  private Random random;  // a fancy random object

  private String outputDirectory;
  // this is where the Text, HTML and XML output as well as the protocol file will be stored

  private String populationFile;
  // a population backup file, which will be used for initialization

  private long startingTime;     // the time when the evolution has been started
  private long lastChange;       // the time when the last improvement of the solution took place

  private static final int OUTPUT_INTERVAL =   20000;   // write protocol information every 20 seconds
  private static final int BACKUP_INTERVAL = 1200000;   // write backup information to disk every 20 minutes


  /** This method is used by the constructors to set the initial parameter values. **/
  private void initializeParameters() {
    random = GlobalObjects.random;
    runMode = GENERATIONAL;
    populationSize = 40;
    crossOverIndividuals = 2;
    crossOverPoints = 0;
    selectionMethod = RANK_BASED_SELECTION;
    scalingMethod = GOLDBERG_SCALING;
    tournamentSize = 2;
    mutationRate = 0.005;
    directedMutationRate = 0;
    alwaysKeepBest = true;
    alwaysReplaceWorst = true;
    elitePercentage = 5;

    // these values will be used in non-hybrid mode to determine when hybrid
    // operators shall be used as part of the canonical evolution process
    repairRate = 0.2;
    climbRate = 0.05;

    generationCount = 0;
    localSearchAfterEvolution = false;
    bestFitness = bestSoFar = Double.NEGATIVE_INFINITY;

    // pre-set termination criteria: stop execution as soon as
    maxGenerationCount = 5000000;  // generation count has reached 5,000,000
    timeLimit = 1800;             // or over-all running time has exceeded 1800 seconds
    stagnationLimit = 300;        // or there has been no improvement during the last 300 seconds

    hybridActivationLimit = 150;  // no improvement for 150 seconds: switch to Hybrid Mode
    hybridActivationTime = -1;  // default: never switch to HYBRID mode without reason


    autoParameters = false;
    outputDirectory = null;
    populationFile = null;

    // initialization of operator probabilities for the steady-state hybrid GA
    hybridCrossoverSuccessRate = 0.8;  // try to initialize with sensible data
    hybridClimbSuccessRate = hybridRepairSuccessRate = 0.1;
    hybridRepairProbability = hybridClimbProbability = 0;

    islandMode = false;
    portNumber = IslandFerry.DEFAULT_FERRY_PORT;
    migrationInterval = 3;
    immigrant = null;
    harborFile = null;
    ferry = null;
    
    outputStream = System.out;  // per default, write all information to STDOUT
  } // end of initializeParameters()

  /**
   * Lets the caller set internal parameters like mutation rate or ELITE size.
   * Parameters are passed by a Map looking like "ELITE_SIZE=30, MUTATION_RATE=0.02".
   **/
  public void setParameters(Map params) {
    if (params == null)
      return;
    Iterator it = ((Set)params.keySet()).iterator();
    while (it.hasNext()) {
      String key = (String)it.next();
      String value = (String)params.get(key);

      // for a description of the parameters, have a look at their declarations above
      try {
        if (key.equals("MUTATION_RATE"))
          mutationRate = Double.parseDouble(value);
        else if (key.equals("DIRECTED_MUTATION_RATE"))
          directedMutationRate = Double.parseDouble(value);
        else if (key.equals("RUN_MODE"))
          runMode = Integer.parseInt(value);
        else if (key.equals("POPULATION_SIZE"))
          populationSize = Integer.parseInt(value);
        else if (key.equals("CROSSOVER_INDIVIDUALS"))
          crossOverIndividuals = Integer.parseInt(value);
        else if (key.equals("CROSSOVER_POINTS"))
          crossOverPoints = Integer.parseInt(value);
        else if (key.equals("SCALING_METHOD"))
          scalingMethod = Integer.parseInt(value);
        else if (key.equals("SELECTION_METHOD"))
          selectionMethod = Integer.parseInt(value);
        else if (key.equals("TOURNAMENT_SIZE"))
          tournamentSize = Integer.parseInt(value);
        else if (key.equals("ELITE_PERCENTAGE"))
          elitePercentage = Integer.parseInt(value);
        else if (key.equals("REPAIR_RATE"))
          repairRate = Double.parseDouble(value);
        else if (key.equals("REPAIR_PROBABILITY"))
          hybridRepairProbability = Double.parseDouble(value);
        else if (key.equals("CLIMB_RATE"))
          climbRate = Double.parseDouble(value);
        else if (key.equals("CLIMB_PROBABILITY"))
          hybridClimbProbability = Double.parseDouble(value);
        else if (key.equals("KEEP_BEST"))
          alwaysKeepBest = Boolean.valueOf(value).booleanValue();
        else if (key.equals("REPLACE_WORST"))
          alwaysReplaceWorst = Boolean.valueOf(value).booleanValue();
        else if (key.equals("LOCAL_SEARCH"))
          localSearchAfterEvolution = Boolean.valueOf(value).booleanValue();
        else if (key.equals("MAX_GENERATION_COUNT"))
          maxGenerationCount = Integer.parseInt(value);
        else if (key.equals("TIME_LIMIT"))
          timeLimit = Integer.parseInt(value);
        else if (key.equals("STAGNATION_LIMIT"))
          stagnationLimit = Integer.parseInt(value);
        else if (key.equals("HYBRID_ACTIVATION_LIMIT"))
          hybridActivationLimit = Integer.parseInt(value);
        else if (key.equals("HYBRID_ACTIVATION_TIME"))
          hybridActivationTime = Integer.parseInt(value);
        else if (key.equals("ISLAND_MODE"))
          islandMode = Boolean.valueOf(value).booleanValue();
        else if (key.equals("HARBOR_FILE"))
          harborFile = value;
        else if (key.equals("PORT_NUMBER"))
          portNumber = Integer.parseInt(value);
        else if (key.equals("MIGRATION_INTERVAL"))
          migrationInterval = Integer.parseInt(value);
        else if (key.equals("OUTPUT")) {
          if (value.endsWith("/"))
            outputDirectory = value;
          else
            outputDirectory = value + "/";
        }
        else if (key.equals("POPULATION_FILE"))
          populationFile = value;
      }
      catch (Exception e) {
        System.err.println("Error setting parameter value: " + e.toString());
      }
    }
  } // end of setParameters(Map)

  /** Alternative method for setting parameters by comma-separated key-value pairs. **/
  public void setParameters(String keyValues) {
    String key, value, keyValue;
    int index = keyValues.indexOf(',');
    Map map = new TreeMap();
    while (index > 0) {
      keyValue = keyValues.substring(0, index).trim();
      keyValues = keyValues.substring(index + 1);
      index = keyValue.indexOf('=');
      if (index > 0) {
        key = keyValue.substring(0, index).trim();
        value = keyValue.substring(index + 1).trim();
        map.put(key, value);
      }
      index = keyValues.indexOf(',');
    }
    index = keyValues.indexOf('=');
    if (index > 0) {
      key = keyValues.substring(0, index).trim();
      value = keyValues.substring(index + 1).trim();
      map.put(key, value);
    }
    setParameters(map);
  } // end of setParameters(String)

  /** This constructor does nothing but initializing the parameters. **/
  public GeneticAlgorithm() {
    initializeParameters();
  }

  /**
   * This constructor is typically called by the scheduling problem that
   * has to be solved. (Since <tt>GeneticAlgorithm</tt> implements the
   * <tt>TTSolver</tt> interface.
   * @param adam   the first schedule (individual). The initial population
   * is created executing subsequent calls to <tt>adam.createNewInstance()</tt>
   * (rib-cutting, if you want so).
   * @param parameters   contains a Map of parameters to be applied to the
   * run of the genetic algorithm (e.g. mutation rate, population size, ...).
   **/
  public GeneticAlgorithm(GeneticIndividual adam, Map parameters) {
    initializeParameters();
    setParameters(parameters);
    population = new GeneticIndividual[populationSize];
    population[0] = adam.duplicate();
    for (int i = 1; i < populationSize; i++) {
      population[i] = adam.duplicate();
      population[i].mutate(1);
    }
    fitnessValues = new double[populationSize];
    calculatePopulationFitness();
  } // end of GeneticAlgorithm(TTProblem, Map)

  /** String-based variant. **/
  public GeneticAlgorithm(GeneticIndividual adam, String parameters) {
    initializeParameters();
    setParameters(parameters);
    population = new GeneticIndividual[populationSize];
    population[0] = adam.duplicate();
    for (int i = 1; i < populationSize; i++) {
      population[i] = adam.duplicate();
      population[i].mutate(1);
    }
    fitnessValues = new double[populationSize];
    calculatePopulationFitness();
  }

  /**
   * Creates a new GeneticAlgorithm instance and sets the initial population to
   * <tt>initialPopulation</tt>. This constructor is used by the first one
   * and also by the <tt>main</tt> method when running the genetic algorithm
   * as stand-alone application (for testing purposes).
   **/
  public GeneticAlgorithm(List initialPopulation) {
    initializeParameters();
    populationSize = initialPopulation.size();
    population = new GeneticIndividual[populationSize];
    for (int i = 0; i < populationSize; i++)
      population[i] = (GeneticIndividual)initialPopulation.get(i);
    fitnessValues = new double[populationSize];
    calculatePopulationFitness();
  } // end of GeneticAlgorithm(List)

  /**
   * <tt>termination</tt> returns true if the termination criteria are
   * fulfilled, else otherwise. In addition, it checks whether time has
   * come to switch to HYBRID mode and does so, if necessary.
   **/
  private boolean termination() {
    long now = System.currentTimeMillis();
    if (generationCount > maxGenerationCount)
      return true;
    if (now - startingTime > timeLimit * 1000)
      return true;
    if (now - lastChange > stagnationLimit * 1000)
      return true;
    if (runMode != HYBRID) {
      if ((hybridActivationLimit >= 0) && (now - lastChange > hybridActivationLimit * 1000)) {
        runMode = HYBRID;
        lastChange = now;
        outputStream.println("[Switched to Hybrid Mode.]");
        writeToLogFile("# Switched to Hybrid Mode.\n");
        savePopulation(outputDirectory + "ga-save.population-switch");
      }
      if ((hybridActivationTime >= 0) && (now - startingTime > hybridActivationTime * 1000)) {
        lastChange = now;
        runMode = HYBRID;
        outputStream.println("[Switched to Hybrid Mode.]");
        writeToLogFile("# Switched to Hybrid Mode.\n");
        savePopulation(outputDirectory + "ga-save.population-switch");
      }
    }
    return false;
  } // end of termination()

  /** Sets the internal mutation rate of the genetic algorithm to <tt>newRate</tt>. **/
  public void setMutationRate(double newRate) {
    if ((newRate >= 0) && (newRate <= 1))
      mutationRate = newRate;
  }

  /** Returns the current mutation rate. **/
  public double getMutationRate() {
    return mutationRate;
  }

  /** Returns the current ELITE size. **/
  public int getEliteSize() {
    return populationSize * elitePercentage / 100;
  }

  /** Returns the current selection method. **/
  public int getSelectionMethod() {
    return selectionMethod;
  }

  /** Sets the output stream, i.e. where on-screen information shall be written to. */
  public void setOutputStream(OutputStream o) {
    outputStream = new PrintStream(o);
  }

  /** Sets the output stream, i.e. where on-screen information shall be written to. */
  public void setOutputStream(PrintStream p) {
    outputStream = p;
  }

  /**
   * This method calculates the fitness of every member of the
   * current population and performs some scaling (removing negative
   * fitness values, e.g.).
   **/
  private void calculatePopulationFitness() {
    int i;
    double avg, min, max;

    // sort the population with respect to the individuals' fitness values.
    // This is necessary for rank-based selection or steady-state GA
    // Sorting is done in ascending order (with respect to fitness values), i.e. population[0]
    // becomes the worst individual.
    
    Arrays.sort(population);

    // now, determine min, max and average fitness values
     
    populationFitness = 0;

    for (i = 0; i < populationSize; i++) {
      fitnessValues[i] = population[i].getFitness();
      populationFitness += fitnessValues[i];
    } // end of for

    previousBestFitness = bestFitness;

    min = fitnessValues[0];
    bestFitness = max = fitnessValues[populationSize - 1];
    GeneticIndividual bestIndividual = population[populationSize - 1];

    if ((bestCreature == null) || (bestFitness > bestCreature.getFitness())) {
      bestCreature = bestIndividual.duplicate();
      lastChange = System.currentTimeMillis();
    }

    // we know different selection methods, so we have to choose
    if (selectionMethod == CANONICAL_SELECTION) {
      double a, b;
      avg = populationFitness / populationSize;
      if (scalingMethod == GOLDBERG_SCALING) {
        // use the scaling algorithm proposed by Goldberg on pp. 77-79
        double mult = 2;
        if (min > (mult * avg - max) / (mult - 1)) {
          a = (mult - 1) * avg / (max - avg);
          b = avg * (max - mult * avg) / (max - avg);
        }
        else {
          a = avg / (avg - min);
          b = -min * avg / (avg - min);
        }
      }
      else if (scalingMethod == SQRT_SCALING) { // this is a scaling algorithm "invented" by me
        a = Math.sqrt(populationSize) / (max - min);
        b = 1 - a * min;
      }
      else { // scalingMethod == DEFAULT_SCALING {
        a = 2 / (max - min);
        b = 1 - a * min;
      }

      scaledFitness = 0;
      for (i = 0; i < populationSize; i++) {
        fitnessValues[i] = a * fitnessValues[i] + b;
        scaledFitness += fitnessValues[i];
      }
    }
  } // end of calculatePopulationFitness()

  /** <tt>selectIndividualIndex</tt> returns the index in the population
    * of the individual selected by the SELECT operator. **/
  private final int selectIndividualIndex() {
    int i = -1;
    if (selectionMethod == CANONICAL_SELECTION) {
      // with CANONICAL_SELECTION, every individual is selected
      // with probability p proportional to its (scaled fitness)
      double selection = random.nextDouble();
      do {
        i++;
        selection -= fitnessValues[i] / scaledFitness;
      } while (selection > 0);
      return i;
    }
    if (selectionMethod == RANK_BASED_SELECTION) {
      // with RANK_BASED_SELECTION, the probability of the best individual
      // to be selected is n/(n*(n+1)/2), of the second best it is
      // (n-1)/(n*(n+1)/2), and so on.
      int selection = random.nextInt(populationSize * (populationSize + 1) / 2) + 1;
      do {
        i++;
        selection -= (i + 1);
      } while (selection > 0);
      return i;
    }
    if (selectionMethod == TOURNAMENT_SELECTION) {
      // Tournament Selection randomly takes n individuals from the population
      // and selects the best (according to fitness value) of these.
      int best = random.nextInt(populationSize);
      for (i = 1; i < tournamentSize; i++) {
        int selection = random.nextInt(populationSize);
        if (fitnessValues[selection] > fitnessValues[best])
          best = selection;
      }
      return best;
    }
    return -1;
  } // end of selectIndividualIndex()

  /**
   * When the Genetic Algorithm runs in island mode (distributed GA), this method
   * can be used by the Ferry to add individuals received from other islands to
   * the current population.
   **/
  protected synchronized final void addIndividual(byte data[]) {
    immigrant = data;
  } // end of addIndividual(GeneticIndividual)

  /**
   * selectIndividual() selects an individual with respect to the
   * selection method that is to be used (e.g. TOURNAMENT_SELECTION)
   **/
  private final GeneticIndividual selectIndividual() {
    int index = selectIndividualIndex();
    if (index < 0)
      return null;
    return population[index];
  } // end of selectIndividual()

  /** Returns the best (fittest) individual of the current population. **/
  public final GeneticIndividual getBestIndividual() {
    int best = 0;
    double bestValue = fitnessValues[0];
    for (int i = 1; i < populationSize; i++)
      if (fitnessValues[i] > bestValue) {
        bestValue = fitnessValues[i];
       best = i;
      }
    return population[best];
  } // end of getBestIndividual()

  /**
   * Returns the historically best individual. That individual does not necessarily
   * have to be part of the current population.
   **/
  public final GeneticIndividual getBestSoFar() {
    return bestCreature;
  }
  
  /**
   * The <tt>evolve()</tt> method creates a new generation from the current one.
   * Different run modes (Generational, Steady-state, Hybrid) are define the actual
   * behaviour of the evolution process.
   **/
  public void evolve() {
    int i, start;
    int newPopulationSize = 0;


    if (runMode == GENERATIONAL) {
      // this is the original GA, proposed by Holland
     
      GeneticIndividual[] newPopulation = new GeneticIndividual[populationSize];

      // perform the ELITE operations (keep best and elite itself)
      if (alwaysKeepBest)
        newPopulation[newPopulationSize++] = getBestIndividual().duplicate();

      while (newPopulationSize < elitePercentage * populationSize / 100) {
        newPopulation[newPopulationSize] = selectIndividual().duplicate();
        newPopulation[newPopulationSize].mutate(mutationRate);
        if (directedMutationRate > 0)
          newPopulation[newPopulationSize].performDirectedMutation(directedMutationRate);
        if ((hybridRepairProbability > 0) && (random.nextDouble() < hybridRepairProbability))
          newPopulation[newPopulationSize].repair(repairRate);
        newPopulation[newPopulationSize].getFitness();
        newPopulationSize++;
      }

      // time for crossover
      while (newPopulationSize < populationSize) {
        GeneticIndividual mother = selectIndividual();
        GeneticIndividual[] fathers = new GeneticIndividual[crossOverIndividuals - 1];
        for (i = 0; i < crossOverIndividuals - 1; i++)
          fathers[i] = selectIndividual();
        GeneticIndividual child = mother.crossOver(fathers, crossOverPoints);
        child.mutate(mutationRate);
        if (directedMutationRate > 0)
          child.performDirectedMutation(directedMutationRate);
        if ((hybridRepairProbability > 0) && (random.nextDouble() < hybridRepairProbability))
          child.repair(repairRate);
        if ((hybridClimbProbability > 0) && (random.nextDouble() < hybridClimbProbability))
          child.climb(climbRate);
        child.getFitness();
        newPopulation[newPopulationSize++] = child;
      } // end of cross-over

      // set current population := new population
      for (i = 0; i < populationSize; i++)
        population[i] = newPopulation[i];
    } // end of runMode == GENERATIONAL


    if (runMode == STEADY_STATE) {  // this is the simple steady-state GA
      GeneticIndividual mother = selectIndividual();
      GeneticIndividual[] fathers = new GeneticIndividual[crossOverIndividuals - 1];
      for (i = 0; i < crossOverIndividuals - 1; i++)
        fathers[i] = selectIndividual();
      GeneticIndividual child = mother.crossOver(fathers, crossOverPoints);
      if (mutationRate > 0)
        child.mutate(mutationRate);
      if (directedMutationRate > 0)
        child.performDirectedMutation(directedMutationRate);
      if (hybridRepairProbability > 0)
        if (random.nextDouble() < hybridRepairProbability)
          child.repair(repairRate);
      if (hybridClimbProbability > 0)
        if (random.nextDouble() < hybridClimbProbability)
          child.repair(climbRate);

      // if the newly-created individual is fitter than the worst one, replace the latter
      if (child.getFitness() > population[0].getFitness())
        population[0] = child;
      if (alwaysReplaceWorst)  // modification to the original steady-state GA: always replace
        population[0] = child;
    } // end of runMode == STEADY_STATE


    if (runMode == HYBRID) {  // hybrid GA with hill-climbing mechanisms

      // the hybrid GA supports four basic genetic operators:
      // cross-over, mutation, hill-climbing and reparation, where
      // cross-over and mutation belong together (atomic). The hybrid GA
      // continuously determines success rates of all operators and sets
      // their operation probability according to these success rates.
      // minimum operator probability is 2.5 percent.

      // recalculate operator probabilities
      double allSum = hybridCrossoverSuccessRate + hybridClimbSuccessRate + hybridRepairSuccessRate;
      if (allSum == 0)
        hybridCrossoverProbability = hybridClimbProbability = hybridRepairProbability = 0.333;
      else {
        hybridCrossoverProbability = 0.025 + 0.925 * (hybridCrossoverSuccessRate / allSum);
        hybridClimbProbability = 0.025 + 0.925 * (hybridClimbSuccessRate / allSum);
        hybridRepairProbability = 0.025 + 0.925 * (hybridRepairSuccessRate / allSum);
      }

      // randomly select the operator to be used, according to calculated probabilities

      double selection = random.nextDouble();
      if (selection < hybridCrossoverProbability) {
        // perform a crossover operation
        GeneticIndividual mother = selectIndividual();
        GeneticIndividual[] fathers = new GeneticIndividual[1];
        fathers[0] = selectIndividual();
        GeneticIndividual child = mother.crossOver(fathers, crossOverPoints);
        if (mutationRate > 0)
          child.mutate(mutationRate);
        if (directedMutationRate > 0)
          child.performDirectedMutation(directedMutationRate);
        hybridCrossoverSuccessRate *= (1 - 1.0/hybridPeriod);
        if (child.getFitness() > population[0].getFitness()) {
          population[0] = child;
          hybridCrossoverSuccessRate += 1.0/hybridPeriod;
        }
        if (alwaysReplaceWorst)  // modification to the original steady-state GA: always replace
          population[0] = child;
      }
      else if (selection < hybridCrossoverProbability + hybridClimbProbability) {
        // perform a hybrid operation (hill-climbing)
//        int index = selectIndividualIndex();
        int index = populationSize - 1 - random.nextInt(populationSize * elitePercentage / 100);
        double oldFitness = population[index].getFitness();
        GeneticIndividual newGI = population[index].duplicate();
        newGI.climb(climbRate);
        hybridClimbSuccessRate *= (1 - 1.0/hybridPeriod);
        if (newGI.getFitness() > oldFitness) {
          population[index] = newGI;
          hybridClimbSuccessRate += 1.0/hybridPeriod;
        }
      }
      else {
        // perform a hybrid operation (repair)
//        int index = selectIndividualIndex();
        int index = populationSize - 1 - random.nextInt(populationSize * elitePercentage / 100);
        GeneticIndividual newGI = population[index].duplicate();
        double oldFitness = newGI.getFitness();
        newGI.repair(repairRate);
        hybridRepairSuccessRate *= (1 - 1.0/hybridPeriod);
        if (newGI.getFitness() > oldFitness) {
          population[index] = newGI;
          hybridRepairSuccessRate += 1.0/hybridPeriod;
        }
      } // end of repair
    } // end of runMode == HYBRID


    calculatePopulationFitness();


    generationCount++;
  } // end of evolve()

  /** Returns a multi-line String description of the current parameter setup. **/
  public String exportParameters() {
    String result = "";
    result += "# RUN_MODE = " +
              (runMode==GENERATIONAL?"GENERATIONAL":(runMode==STEADY_STATE?"STEADY_STATE":"HYBRID")) + ".\n";
    result += "# POPULATION_SIZE = " + populationSize + ".\n";
    result += "# KEEP_BEST = " + alwaysKeepBest + ".\n";
    result += "# LOCAL_SEARCH = " + localSearchAfterEvolution + ".\n";
    if (runMode != HYBRID) {  // in hybrid mode (MODE 3), reparation cannot be turned off
      if (hybridRepairProbability == 0)
        result += "# REPAIR_PROBABILITY = 0.\n";
      else
        result += "# REPAIR_PROBABILITY = " + hybridRepairProbability +
                  " (REPAIR_RATE = " + repairRate + ").\n";
    }
    else
      result += "# REPAIR_RATE = " + repairRate + ".\n";
    result += "# ELITE_PERCENTAGE = " + elitePercentage + ".\n";
    result += "# MUTATION_RATE = " + mutationRate + ".\n";
    result += "# DIRECTED_MUTATION_RATE = " + directedMutationRate + ".\n";
    result += "# CROSSOVER_INDIVIDUALS = " + crossOverIndividuals + ".\n";
    result += "# CROSSOVER_POINTS = " + crossOverPoints + ".\n";
    result += "# SELECTION_METHOD = ";
    if (selectionMethod == CANONICAL_SELECTION) {
      result += "CANONICAL_SELECTION.\n";
      result += "# SCALING_METHOD = ";
      if (scalingMethod == GOLDBERG_SCALING)
        result += "Goldberg's Method.\n";
      else if (scalingMethod == SQRT_SCALING)
        result += "Buettcher's SQRT Method.\n";
      else
        result += "Default Scaling (x2).\n";
    }
    else if (selectionMethod == TOURNAMENT_SELECTION) {
      result += "TOURNAMENT_SELECTION.\n";
      result += "# TOURNAMENT_SIZE = " + tournamentSize + ".\n";
    }
    else
      result += "RANK_BASED_SELECTION.\n";

    result += "# MAX_GENERATION_COUNT = " + maxGenerationCount + ".\n";
    result += "# TIME_LIMIT = " + timeLimit + " seconds.\n";
    result += "# STAGNATION_LIMIT = " + stagnationLimit + " seconds.\n";
    if (runMode != HYBRID) {
      result += "# HYBRID_ACTIVATION_LIMIT = " + hybridActivationLimit + " seconds.\n";
      if (hybridActivationTime < 0)
        result += "# HYBRID_ACTIVATION_TIME = never\n";
      else
        result += "# HYBRID_ACTIVATION_TIME = " + hybridActivationTime + " seconds\n";
    }

    result += "# \n";
    if (outputDirectory == null) {
      result += "# Output Directory: [Unspecified]\n";
      result += "# Warning: No Output Directory specified!\n";
    }
    else
      result += "# Output Directory: " + outputDirectory + "\n";

    return result;
  } // end of exportParameters()

  /**
   * <tt>autoSetup</tt> sets some of the internal parameters according to the
   * result of a training phase.
   **/
  public void autoSetup(int timeLimit) {

    int bestSelectionMethod, bestScalingMethod, seM, scM;
    double bestMutationRate, muR;
    double bestResult = Double.NEGATIVE_INFINITY;

//    GAThread threads[] = new GAThreads[24];
    for (int i = 0; i < 24; i++) {
    }
//    0.001 0.002 0.005 0.01 0.02 0.03 0.05 0.08
    outputStream.println("[AUTOSETUP]");
    outputStream.print(exportParameters());
  } // end of autoSetup(int, int)

  /**
   * The <tt>run()</tt> method calls <tt>evolve</tt> (each call generates one
   * new generation) until the termination criterion is fulfilled.
   **/
  public GeneticIndividual run() {

    if (islandMode) {  // distributed GA run
      ferry = new IslandFerry(this, null, portNumber);  // just create a new ferry for GI transportation
      if (harborFile != null)
        ferry.addHarborFile(harborFile);
    }

    lastChange = startingTime = System.currentTimeMillis();
    long lastOutputTime = System.currentTimeMillis();
    long lastBackupTime = System.currentTimeMillis();

    // create the protocol file now
    if (outputDirectory != null) try {
      FileWriter fw = new FileWriter(outputDirectory + "ga.log");
      String newString = exportParameters() + "\n\n";
      fw.write(newString.toCharArray(), 0, newString.length());
      // initialize column names
      newString = "Generation\tBest_Fitness\tAverage_Fitness\tTime_elapsed";
      if (runMode == HYBRID)  // hybrid mode
        newString += "\tCrossover_Probability\tRepair_Probability\tHill-Climbing_Probability";
      List individualInfo = population[0].getProtocolInformation();
      for (int individualCnt = 0; individualCnt < individualInfo.size(); individualCnt++)
        newString += "\t" + GlobalMethods.replaceString(((KeyValue)individualInfo.get(individualCnt)).key, " ", "_");
      newString += "\n";
      fw.write(newString.toCharArray(), 0, newString.length());
      fw.close();
    } catch (IOException ioex) { }

    startMode = runMode;

    if (populationFile != null) {
      outputStream.println("Reading initial population from file: " + populationFile + "\n");
      restorePopulation(populationFile);
    }

    outputStream.println("\n\nStarting Genetic Algorithm. Parameter Setup:");
    outputStream.println(exportParameters());  // tell the user what parameters are to be used

    outputStream.println("\nEvolving [" + population[0].getSpeciesDescription() + "].\n");

    // this DecimalFormat object will be used to write fitness values, etc.
    // to stdout and the protocol file
    DecimalFormat decFormat = new DecimalFormat("0.000");

    int migrationCountDown = migrationInterval;
    if (runMode != GENERATIONAL)  // steady-state GA: scale migrationFrequency
      migrationCountDown *= populationSize;

    do {
      // first, create a new population
      evolve();

      // The following section manages the Island Mode (distributed GA run).
      // In Island Mode, the GA runs on a number of distinct machines (islands).
      // Every 10 seconds, every island sends an individual -- selected by the
      // selectIndividual() method -- to another, randomly chosen, island.

      migrationCountDown--;
      if (migrationCountDown == 0) {
        // send a randomly-chosen individual to another population (island)
        if (islandMode) {  // in case of distributed GA run (island mode),
                           // send an individual to a randomly-chosen island
          GeneticIndividual traveller = selectIndividual();
          ferry.sendIndividual(traveller.getGeneticCode(), ferry.randomHarbor());
        }

        migrationCountDown = migrationInterval;
        if (runMode != GENERATIONAL)
          migrationCountDown *= populationSize;
      } // end of migration

      if (islandMode) {  // check for new individual from another island
        byte newIndividual[] = immigrant;
        immigrant = null;
        if (newIndividual != null) {  // a new individual has arrived
          population[0].setGeneticCode(newIndividual);
          outputStream.println("New individual's fitness: " + decFormat.format(population[0].getFitness()));
          calculatePopulationFitness();
        }
      }

      long now = System.currentTimeMillis();
      if (now - lastOutputTime > OUTPUT_INTERVAL) {
        // print some evolutionary information to screen and log file

        outputStream.println("Generation #" + generationCount + "\t" +
                           ((System.currentTimeMillis() - startingTime) / 1000) + " seconds\t" +
                           decFormat.format(bestFitness));

        // write protocol information to file
        if (outputDirectory != null) try {
          FileWriter fw = new FileWriter(outputDirectory + "ga.log", true);
          StringBuffer formattedNumber = new StringBuffer();
          String newString = "" + generationCount + "\t" +
                                  decFormat.format(bestFitness) + "\t" +
                                  decFormat.format(populationFitness/populationSize) + "\t" +
                                  ((System.currentTimeMillis() - startingTime) / 1000);
          if (startMode == HYBRID)  // hybrid mode
            newString += "\t" + decFormat.format(hybridCrossoverProbability) +
                         "\t" + decFormat.format(hybridRepairProbability) +
                         "\t" + decFormat.format(hybridClimbProbability);

          List individualInfo = bestCreature.getProtocolInformation();

          for (int individualCnt = 0; individualCnt < individualInfo.size(); individualCnt++)
            newString += "\t" + ((KeyValue)individualInfo.get(individualCnt)).value.toString();
          newString += "\n";
          fw.write(newString.toCharArray(), 0, newString.length());
          fw.close();
        } catch (IOException ioe) {
          System.err.println("Error: Could not write to protocol file.");
        }

        // adjust the clock
        lastOutputTime = System.currentTimeMillis();

      }  // end of protocol file output

      try {
        if (System.in.available() > 0) {
          BufferedReader r = new BufferedReader(new InputStreamReader(System.in));
          String s = r.readLine();
          if (s.compareTo("STOP") == 0)
            break;
          else
            setParameters(s);
        }
      }
      catch (IOException ioe) { }

      if (now - lastBackupTime > BACKUP_INTERVAL) {  // every N minutes, save the best individual and the current population
        if (outputDirectory != null) {
          outputStream.println("[Saving best individual.]");
          saveToDisk(bestCreature.getSummary(), outputDirectory + "ga-save.txt");
          saveToDisk(bestCreature.toHTML(), outputDirectory + "ga-save.html");
          saveToDisk(bestCreature.toXML(), outputDirectory + "ga-save.xml");
          outputStream.println("[Saving current population.]");
          savePopulation(outputDirectory + "ga-save.population");
        }
        lastBackupTime = now;
      } // end of backup stuff

    } while (!termination());

    // evolution is over, so stop any ferry activities immediately
    if (ferry != null) {
      ferry.sink();
      ferry = null;
    }

    outputStream.println("Genetic Algorithm finished.");
    outputStream.println("Final Fitness Value: " + bestCreature.getFitness());

    if (localSearchAfterEvolution) {
      bestCreature.localSearch();
      outputStream.println("After Local Search: " + bestCreature.getFitness());
    }

    if (outputDirectory != null) {
      outputStream.print("Saving final outcome: ");
      saveToDisk(bestCreature.getSummary(), outputDirectory + "ga.txt");
      outputStream.print("txt");
      saveToDisk(bestCreature.toHTML(), outputDirectory + "ga.html");
      outputStream.print(", html");
      saveToDisk(bestCreature.toXML(), outputDirectory + "ga.xml");
      outputStream.println(", xml.");
    }

    outputStream.println("Execution finished.");

    return bestCreature;
  } // end of run()

  /**
   * Since <tt>GeneticAlgorithm</tt> implements <tt>TTSolver</tt>, it needs
   * a <tt>solve</tt> method. This method is just a wrapper function that
   * calls <tt>run</tt> and then returns the result of that call.
   **/
  public TTProblem solve() {
    TTChromosome outcome = (TTChromosome)run();
    outcome.fitness = GeneticIndividual.NO_FITNESS;
    outcome.getFitness();
    return outcome.getTTProblem();
  }

  /** Inherited from TTSolver **/
  public void setProblem(TTProblem problem) {
    population = new GeneticIndividual[populationSize];
    population[0] = new TTChromosome(problem);
    for (int i = 1; i < populationSize; i++) {
      population[i] = population[0].duplicate();
      population[i].mutate(1);
    }
    fitnessValues = new double[populationSize];
    calculatePopulationFitness();
  }

  /** Returns the current generation number. **/
  public int getGenerationCount() {
    return generationCount;
  }

  /** Returns the run mode (GENERATIONAL, STEADY_STATE or HYBRID). **/
  public int getRunMode() {
    return runMode;
  }
  
  /** Returns the fitness of the best individual. **/
  public double getBestFitness() {
    return bestFitness;
  }
  
  /** Returns the average fitness value of the current population. **/
  public double getAverageFitness() {
    return populationFitness / populationSize;
  }

  /** Writes <tt>s</tt> to the log file, if output directory specified. **/
  private void writeToLogFile(String s) {
    if (outputDirectory == null) return;
    try {
      FileWriter fw = new FileWriter(outputDirectory + "ga.log", true);
      fw.write(s.toCharArray(), 0, s.length());
      fw.close();
    }
    catch (IOException ioe) {
      System.err.println("Error writing to log file: " + ioe.toString());
    }
  } // end of writeToLogFile(String)

  /**
   * Writes the String <tt>individual</tt> to the file specified by
   * <tt>fileName</tt> previous content is destroyed.
   **/
  private void saveToDisk(String individual, String fileName) {
    try {
      OutputStream os = new FileOutputStream(fileName);
      os.write(individual.getBytes());
      os.flush();
      os.close();
    } catch (IOException ioe) {
      System.err.println("\nAn IOException occurred while trying to save individual.\n");
      System.err.println(ioe.toString());
    }
  } // end of saveToDisk(String, String)

  /** Transforms an integer value into a byte array (useful when writing ints to a byte stream). **/
  private static byte[] intToByteArray(int value) {
    byte result[] = new byte[4];
    for (int i = 3; i >= 0; i--) {
      result[i] = (byte)(value % 120);
      value /= 120;
    }
    return result;
  } // end of intToByteArray(int)

  /** Transforms a byte array into an integer (useful when reading byte input streams, which contain integer data). **/
  private static int byteArrayToInt(byte value[]) {
    int result = 0;
    for (int i = 0; i < value.length; i++) {
      result *= 120;
      result += value[i];
    }
    return result;
  } // end of byteArrayToInt(byte[])

  /** Saves the current population to a backup file. **/
  private void savePopulation(String fileName) {
    try {
      int i;
      byte header[] = new byte[3];
      header[0] = 'G'; header[1] = 'A'; header[2] = 0;
      byte generation[] = intToByteArray(generationCount);
      byte time[] = intToByteArray((int)(System.currentTimeMillis() - startingTime));
      byte size[] = intToByteArray(populationSize);
      OutputStream os = new FileOutputStream(fileName);
      os.write(header);
      os.write(generation);
      os.write(time);
      os.write(size);
      for (i = 0; i < populationSize; i++) {
        byte data[] = population[i].getGeneticCode();
        size = intToByteArray(data.length);
        os.write(size);
        os.write(data);
      }
      os.flush();
      os.close();
    }
    catch (IOException ioe) {
      System.err.println("IOException while trying to save current population.");
      System.err.println(ioe.toString());
    }
  } // end of savePopulation

  /** Restores the population using a previously created backup file. **/
  private void restorePopulation(String fileName) {
    try {
      int i, chunkSize;
      byte header[] = new byte[3];
      byte generation[] = new byte[4];
      byte time[] = new byte[4];
      byte size[] = new byte[4];
      InputStream is = new FileInputStream(fileName);
      is.read(header);
      if ((header[0] != 'G') || (header[1] != 'A') || (header[2] != 0)) {
        System.err.println("Trying to restore population from a corrupt backup file.");
        return;
      }
      is.read(generation);
      generationCount = byteArrayToInt(generation);
      is.read(time);
      startingTime = System.currentTimeMillis() - byteArrayToInt(time);
      is.read(size);
      populationSize = byteArrayToInt(size);
      GeneticIndividual adam = population[0];
      population = new GeneticIndividual[populationSize];
      for (i = 0; i < populationSize; i++) {
        population[i] = adam.duplicate();
        is.read(size);
        chunkSize = byteArrayToInt(size);
        byte data[] = new byte[chunkSize];
        is.read(data);
        population[i].setGeneticCode(data);
      }
      fitnessValues = new double[populationSize];
      calculatePopulationFitness();
    }
    catch (IOException ioe) {
      System.err.println("IOException while trying to restore population.");
      System.err.println(ioe.toString());
    }
  } // end of restorePopulation

  /**
   * The <tt>main</tt> method lets the user choose a specific
   * species, which will be created by <tt>God</tt>, and lets
   * the Genetic Algorithm run on an initial population of that
   * type (initial population created by <tt>God</tt>).
   **/
  public static void main(String args[]) {
    String selection, dataFile, input;
    try {
      System.out.print("Please choose the species to evolve: ");
      if (args.length >= 1) {
        selection = args[0];
        System.out.println("" + selection);
      }
      else
        selection = new BufferedReader(new InputStreamReader(System.in)).readLine();

      System.out.print("Please define the initialization file: ");
      if (args.length >= 2) {
        dataFile = args[1];
        System.out.println("" + dataFile);
      }
      else
        dataFile = new BufferedReader(new InputStreamReader(System.in)).readLine();

      System.out.print("Please enter the GA parameters: ");
      if (args.length >= 3) {
        input = args[2];
        System.out.println(input);
      }
      else
        input = new BufferedReader(new InputStreamReader(System.in)).readLine();

      GeneticIndividual solution;
      GeneticIndividual adam = God.createIndividual(selection, dataFile);
      adam.getFitness();

      if (adam != null) {
        GeneticAlgorithm ga = new GeneticAlgorithm(adam, input);
        solution = ga.run();
      }

    }
    catch (IOException ioe) {
      System.err.println("An I/O Error occurred.");
      return;
    }

  } // end of main(String[])

} // end of class GeneticAlgorithm

