/**
 * Implementation of class "IslandFerry".
 *
 * Author: Stefan Buettcher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 **/

package ga;

import helperclasses.*;

import java.io.*;
import java.net.*;
import java.util.*;

/*****
 * The <tt>IslandFerry</tt> class manages transport of individuals from one island
 * to another. This is useful when implementing a distributed Genetic Algorithm.
 *****/

public class IslandFerry {
  public static final int DEFAULT_FERRY_PORT = 9253;

  private GeneticAlgorithm owner;
  private String name;
  private int port;

  /** The listening Datagram socket. **/
  protected DatagramSocket listener;
  /** The packet where the data received will be stored. **/
  protected DatagramPacket inData;

  /** Datagram socket to send data to another GA node. **/
  protected DatagramSocket sender;
  /** The packet that contains the data to be sent. **/
  protected DatagramPacket outData;

  private ArrayList harbors;
  private FerryThread thread;


  /**
   * The inner class <tt>Host</tt> is used for specifying remote hosts by their
   * name and port number.
   **/
  class Host {
    public String name;
    public String hostName;
    public int port;

    public Host(String name, String hostName, int port) {
      this.name = name;
      this.hostName = hostName;
      this.port = port;
    }
    
    public final String toString() {
      return name + " (" + hostName + ":" + port + ")";
    }

  } // end of inner class Host

  /**
   * Constructs a new ferry.
   * @param name  the name of the island the ferry belongs to.
   **/
  public IslandFerry(GeneticAlgorithm owner, String name, int portNumber) {
    this.owner = owner;
    this.name = name;
    this.port = portNumber;

    if (name == null) {  // auto-detect
      try {
        InetAddress self = InetAddress.getLocalHost();
        this.name = self.getHostName();
      }
      catch (UnknownHostException uhe) {
        System.err.println("Cannot determine local host name.");
      }
    }
    System.out.println("Island Ferry is running for harbor \"" + this.name + "\".");

    thread = null;
    try {
      listener = new DatagramSocket(portNumber);
      thread = new FerryThread(this);
      thread.start();
    } catch (SocketException se) {
      if (thread != null)
        thread.destroy();
      thread = null;
      System.err.println("Cannot setup listen socket.");
    }
    harbors = new ArrayList();
  } // end of IslandFerry(String)

  public void addHarbor(String name, String hostName, int port) {
    harbors.add(new Host(name, hostName, port));
  } // end of addHarbor(String)

  public void addHarborFile(String filename) {
    try {
      FileReader fr = new FileReader(filename);
      BufferedReader br = new BufferedReader(fr);
      String line = br.readLine();
      while (line != null) {
        List list = GlobalMethods.listFromString(line, ',');
        try {
          String name = (String)list.get(0);
          String hostName = (String)list.get(1);
          int port = Integer.parseInt((String)list.get(2));
          harbors.add(new Host(name, hostName, port));
        }
        catch (Exception e) {
        }
        line = br.readLine();
      }
    } catch (IOException ioe) {
    }
    System.out.println("Harbor/Island List: " + harbors + ".");
  } // end of addHarborFile(String)

  private Host getHostByName(String name) {
    for (int i = 0; i < harbors.size(); i++) {
      Host host = (Host)harbors.get(i);
      if (host.name.equals(name))
        return host;
    }
    return null;
  } // end of getHostByName(String)

  /** Randomly chooses a harbor from the list of known harbors. **/
  final protected String randomHarbor() {
    if (harbors.size() == 0)
      return null;
    Host host = (Host)harbors.get(GlobalObjects.random.nextInt(harbors.size()));
    return host.name;
  }

  /** Transmits a given individual to a certain destination harbor (GA instance). **/
  public void sendIndividual(byte byteData[], String destination) {
    if (destination == null) {
      System.out.println("Sending individual to nowhere.");
      return;
    }
    Host harbor = getHostByName(destination);
    if (harbor != null) {
      try {
        byte data[] = new byte[byteData.length];
        System.arraycopy(byteData, 0, data, 0, data.length);
        sender = new DatagramSocket();
        InetAddress address = InetAddress.getByName(harbor.hostName);
        outData = new DatagramPacket(data, data.length, address, harbor.port);
        sender.send(outData);
//        System.out.println("Individual sent to \"" + destination + "\" (" + harbor.hostName +
//	                   ", size: " + data.length + " bytes).");
      }
      catch (IOException ioe) {
        System.err.println("Cannot send individual to \"" + destination + "\": IOException.");
        System.err.println(ioe.toString());
        ioe.printStackTrace();
      }
    }
    else
      System.err.println("Destination harbor \"" + destination + "\" does not exist.");
  } // end of sendIndividual(GeneticIndividual, String)

  /**
   * Welcome an individual just arrived on the island and sends it directly to the
   * evolution process (Genetic Algorithm).
   **/
  public synchronized void welcomeNewIndividual(byte data[], String source) {
    System.out.println("New individual received from Island/Host \"" + source + "\".");
    owner.addIndividual(data);
  } // end of welcomeNewIndividual(GeneticIndividual)

  /**
   * This methods lets the ferry sink. After completion of <tt>sink</tt>,
   * the ferry is no longer working (since the listening thread has been
   * destroyed.
   **/
  public void sink() {
    if (thread != null) {
      thread.terminate();
      thread = null;
    }
  } // end of sink()

  /** For test purposes only. **/
  public static void main(String args[]) {
    IslandFerry ferry = new IslandFerry(null, null, DEFAULT_FERRY_PORT);
    try { Thread.sleep(5000); }
      catch (Exception e) { }
    ferry.sink();
  } // end of main(String[])

} // end of class IslandFerry


/*****
 * This class manages listening for incoming individuals.
 *****/

class FerryThread extends Thread {
  private static final int IN_BUFFER_LENGTH = 128 * 1024;

  private IslandFerry owner;
  private boolean terminated;

  public FerryThread(IslandFerry owner) {
    this.owner = owner;
    terminated = false;
  }

  /** The standard <tt>run</tt> method, executing an infinite receive loop. **/
  public void run() {
    do {
      try {
        byte inBuffer[] = new byte[IN_BUFFER_LENGTH];
        owner.inData = new DatagramPacket(inBuffer, IN_BUFFER_LENGTH);
        owner.listener.receive(owner.inData);
        byte data[] = new byte[owner.inData.getLength()];
        System.arraycopy(owner.inData.getData(), 0, data, 0, data.length);
        if ((data != null) && (data.length > 1))  // something received: process the data
          owner.welcomeNewIndividual(data, owner.inData.getAddress().getHostName());
      } catch (IOException ioe) {
        System.err.println("Error receiving individual.");
      }
    } while (!terminated);
  } // end of run()

  public void terminate() {
    terminated = true;
  }

} // end of class FerryThread

