/*****
 * <tt>EfficientArrayList</tt> is just a stupid copy of the <tt>ArrayList</tt>
 * class but makes more efficient use of memory.
 *****/
package helperclasses;

import java.util.*;

public class EfficientArrayList extends AbstractList implements List, Cloneable, java.io.Serializable {
  private transient Object elementData[];
  private int size;

  public EfficientArrayList(int initialCapacity) {
    super();
    if (initialCapacity < 0)
      throw new IllegalArgumentException("Illegal Capacity: " + initialCapacity);
    this.elementData = new Object[initialCapacity];
  }

  public EfficientArrayList() {
    this(4);
  }

  public EfficientArrayList(Collection c) {
    size = c.size();
    elementData = new Object[(size*110)/100]; // Allow 10% room for growth
    c.toArray(elementData);
  }

  public void trimToSize() {
    modCount++;
    int oldCapacity = elementData.length;
    if (size < oldCapacity) {
      Object oldData[] = elementData;
      elementData = new Object[size];
      System.arraycopy(oldData, 0, elementData, 0, size);
    }
  }

  public void ensureCapacity(int minCapacity) {
    modCount++;
    int oldCapacity = elementData.length;
    if (minCapacity > oldCapacity) {
      Object oldData[] = elementData;
      int newCapacity = oldCapacity + 4;
      if (newCapacity < minCapacity)
        newCapacity = minCapacity;
      elementData = new Object[newCapacity];
      System.arraycopy(oldData, 0, elementData, 0, size);
    }
  }
    
  public int size() {
    return size;
  }

  public boolean isEmpty() {
    return size == 0;
  }

  public boolean contains(Object elem) {
    return indexOf(elem) >= 0;
  }

    public int indexOf(Object elem) {
	if (elem == null) {
	    for (int i = 0; i < size; i++)
		if (elementData[i]==null)
		    return i;
	} else {
	    for (int i = 0; i < size; i++)
		if (elem.equals(elementData[i]))
		    return i;
	}
	return -1;
    }

    public int lastIndexOf(Object elem) {
	if (elem == null) {
	    for (int i = size-1; i >= 0; i--)
		if (elementData[i]==null)
		    return i;
	} else {
	    for (int i = size-1; i >= 0; i--)
		if (elem.equals(elementData[i]))
		    return i;
	}
	return -1;
    }

  public Object clone() {
    throw new InternalError();
  }

  public Object[] toArray() {
    Object[] result = new Object[size];
    System.arraycopy(elementData, 0, result, 0, size);
    return result;
  }

    public Object[] toArray(Object a[]) {
        if (a.length < size)
            a = (Object[])java.lang.reflect.Array.newInstance(
                                a.getClass().getComponentType(), size);

	System.arraycopy(elementData, 0, a, 0, size);

        if (a.length > size)
            a[size] = null;

        return a;
    }

    public Object get(int index) {
	RangeCheck(index);

	return elementData[index];
    }

    public Object set(int index, Object element) {
	RangeCheck(index);

	Object oldValue = elementData[index];
	elementData[index] = element;
	return oldValue;
    }

    public boolean add(Object o) {
	ensureCapacity(size + 1);  // Increments modCount!!
	elementData[size++] = o;
	return true;
    }

    public void add(int index, Object element) {
	if (index > size || index < 0)
	    throw new IndexOutOfBoundsException(
		"Index: "+index+", Size: "+size);

	ensureCapacity(size+1);  // Increments modCount!!
	System.arraycopy(elementData, index, elementData, index + 1,
			 size - index);
	elementData[index] = element;
	size++;
    }

    public Object remove(int index) {
	RangeCheck(index);

	modCount++;
	Object oldValue = elementData[index];

	int numMoved = size - index - 1;
	if (numMoved > 0)
	    System.arraycopy(elementData, index+1, elementData, index,
			     numMoved);
	elementData[--size] = null; // Let gc do its work

	return oldValue;
    }

    public void clear() {
	modCount++;

	for (int i = 0; i < size; i++)
	    elementData[i] = null;

	size = 0;
    }

    public boolean addAll(Collection c) {
	modCount++;
	int numNew = c.size();
	ensureCapacity(size + numNew);

	Iterator e = c.iterator();
	for (int i=0; i<numNew; i++)
	    elementData[size++] = e.next();

	return numNew != 0;
    }

    public boolean addAll(int index, Collection c) {
	if (index > size || index < 0)
	    throw new IndexOutOfBoundsException(
		"Index: "+index+", Size: "+size);

	int numNew = c.size();
	ensureCapacity(size + numNew);  // Increments modCount!!

	int numMoved = size - index;
	if (numMoved > 0)
	    System.arraycopy(elementData, index, elementData, index + numNew,
			     numMoved);

	Iterator e = c.iterator();
	for (int i=0; i<numNew; i++)
	    elementData[index++] = e.next();

	size += numNew;
	return numNew != 0;
    }

    protected void removeRange(int fromIndex, int toIndex) {
	modCount++;
	int numMoved = size - toIndex;
        System.arraycopy(elementData, toIndex, elementData, fromIndex, numMoved);
	int newSize = size - (toIndex-fromIndex);
	while (size != newSize)
	    elementData[--size] = null;
    }

    private void RangeCheck(int index) {
	if (index >= size || index < 0)
	    throw new IndexOutOfBoundsException(
		"Index: "+index+", Size: "+size);
    }

    private synchronized void writeObject(java.io.ObjectOutputStream s)
        throws java.io.IOException{
	s.defaultWriteObject();
        s.writeInt(elementData.length);
	for (int i=0; i<size; i++)
            s.writeObject(elementData[i]);
    }

    private synchronized void readObject(java.io.ObjectInputStream s)
        throws java.io.IOException, ClassNotFoundException {
	s.defaultReadObject();
        int arrayLength = s.readInt();
        elementData = new Object[arrayLength];
	for (int i=0; i<size; i++)
            elementData[i] = s.readObject();
    }

} // end of class EfficientArrayList
