/**
 * Implementation of class "SimpleSearch".
 *
 * Author: Stefan Buettcher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 **/

package simplesearch;

import helperclasses.*;
import ttclasses.*;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;

/*****
 * This is just another approach to the Timetabling problem. Its sole purpose is to
 * gain some reference experience that can be used to judge the performance of the
 * Genetic Algorithm.
 *****/
public class SimpleSearch implements TTSolver {
  private TTProblem problem;
  private Random random;

  /** Creates new SimpleSearch */
  public SimpleSearch(TTProblem instance) {
    problem = instance;
    random = GlobalObjects.random;
  }

  private List mixList(List l) {
    List result = new ArrayList();
    int size = l.size();
    while (size > 0) {
      Object element = l.get(random.nextInt(size--));
      l.remove(element);
      result.add(element);
    }
    return result;
  } // end of mixList(List)

  /** This method calculates the total penalty of the Resource/ResourceList pair given. **/
  private double calculatePenalty(PlanningClass pc, boolean onlyHardConstraints) {
    int j;
    double result = 0;
    List l = pc.computeConstraintViolations();
    for (j = 0; j < l.size(); j++) {
      ConstraintViolation cv = (ConstraintViolation)l.get(j);
      result += cv.penalty;
    }
    return result;
  } // end of calculatePenalty(ResourceList, Resource)

  private void unplanResourceList(ResourceList rl) {
    // in case we were not able to plan this
    int j, k;
    List l;
    l = rl.getAssignmentList();
    for (k = 0; k < l.size(); k++)
      rl.freeResource((Resource)l.get(k));
  } // end of unplanEvent(Event)

  private boolean planEvent(Event e) {
    int j, k;
    ResourceList rl;
    List l;
    double penalty1, penalty2;
    boolean successful = true;
    // go through all resource lists of this event
    for (j = 0; j < e.resources.size(); j++) {
      rl = (ResourceList)e.resources.get(j);
      // Min = Max is exactly the number of available resources -> assign all
      if ((rl.getMin() == rl.getMax()) && (rl.getMin() == rl.size())) {
        for (k = 0; k < rl.size(); k++)
          rl.assignResource(k);
      }
      else {
        // Min != Max or number of available resources is more or less than Min/Max
        while (rl.assignedCount() < rl.getMin()) {
          successful = false;
          // Compute randomly a new sequence of the resources available
          l = mixList(rl.getFreeResources());
          for (k = 0; k < l.size(); k++) {
            // this for loop cycles through all resource components possibly matching the criteria
            Resource res = (Resource)l.get(k);
            penalty1 = calculatePenalty(e, true) + calculatePenalty(res, true);
            rl.assignResource(res);
            penalty2 = calculatePenalty(e, true) + calculatePenalty(res, true);
            // if assignment causes higher penalty -> free resource
            if (penalty2 > penalty1)
              rl.freeResource(res);
            else {
              successful = true;
              break;
            }
          } // end of for through components of current type
          if (!successful)
            break;
        }
      }
      if (!successful) {
        unplanResourceList(rl);
        break;
      }
    } // end of loop through resource lists
//    if (!successful)
//      unplanEvent(e);
    return successful;
  } // end of planEvent(Event)

  /** finetuneEvent manages the fine-tuning after the initial planning phase is over. **/
  private void finetuneEvent(Event e) {
    int j, k;
    double penalty1, penalty2;
    // go through all available resources
    for (j = 0; j < e.resources.size(); j++) {
      ResourceList rl = (ResourceList)e.resources.get(j);
      if (rl.assignedCount() < rl.getMax()) {
        // get a list of resources not yet assigned
        List free = mixList(rl.getFreeResources());
        for (k = 0; k < free.size(); k++) {
          Resource res = (Resource)free.get(k);
          penalty1 = calculatePenalty(e, false) + calculatePenalty(res, false);
          rl.assignResource(res);
          penalty2 = calculatePenalty(e, false) + calculatePenalty(res, false);
          if (penalty2 > penalty1)
            rl.freeResource(res);
          if (rl.assignedCount() >= rl.getMax())
            break;
        }
      }
    }
  } // end of finetuneEvent(Event)

  public void setProblem(TTProblem problem) {
    this.problem = problem;
  }

  /**
   * Tries to get a feasible solution of the Timetabling problem.
   **/
  public TTProblem solve() {
    int i, size, cnt;
    int plannedCount = 0;
    Event e;
    boolean success;
    boolean hasBeenPlanned[];
    List events = problem.getEvents();
    hasBeenPlanned = new boolean[events.size()];
    System.out.println("Planning events (first phase).");
    size = events.size();
    for (i = 0; i < size; i++) {
      e = (Event)events.get(i);
      hasBeenPlanned[i] = success = planEvent(e);
      if (success) plannedCount++;
      System.out.print("[" + (i+1) + "/" + size + ": ");
      if (success)
        System.out.print(" Success.]");
      else
        System.out.print(" Failed.]");
      if (i < size-1)
        System.out.println(",");
      else
        System.out.println(".");
    }
    System.out.println("Fine-tuning events (second phase).");
    cnt = 0;
    for (i = 0; i < size; i++) if (hasBeenPlanned[i]) {
      e = (Event)events.get(i);
      finetuneEvent(e);
      System.out.print("[" + (++cnt) + "/" + plannedCount + "]");
      if (cnt < plannedCount)
        System.out.print(", ");
      else
        System.out.println(".");
    }
    System.out.println("Exchanging resources (third phase).");
    return problem;
  } // end of solve()

  public static void main(String args[]) {
    TTProblem problem = null;

    if(args.length < 2) {
      System.out.println("Syntax: java SimpleSearch Problem Datafile.xml");
      return;
    }

    problem = ((TTChromosome)God.createIndividual(args[0], args[1])).getTTProblem();

    System.out.println("[" + problem.getSpeciesDescription() + "]");

    long now = System.currentTimeMillis();
    SimpleSearch bb = new SimpleSearch(problem);
    TTProblem ttp = (TTProblem)bb.solve();
    ttp.getFitness();

    System.out.print("[Saving final outcome: ");
    TTProblem.saveToDisk(ttp.getSummary(), "search.txt");
    System.out.print("txt");
    TTProblem.saveToDisk(ttp.toHTML(), "search.html");
    System.out.print(", html");
    TTProblem.saveToDisk(ttp.toXML(), "search.xml");
    System.out.println(", xml.]");


    System.out.println("\nPlanning finished.\n");
    System.out.println("\n" + ttp.getSummary() + "\n");
    System.out.println("Time elapsed: " + (System.currentTimeMillis() - now) + " ms.\n");
  } // end of main(String[])

} // end of class SimpleSearch
