package tools;

import helperclasses.*;
import java.io.*;
import java.util.*;

/*****
 * The class <tt>ExtractColumns</tt> may be used to extract certain columns from a GA protocol
 * file. The protocol file contains many different aspects of the GA run. This class is useful for
 * extracting only the information that is currently of interest. <br><br>
 * <b>Example.</b> The Genetic Algorithm could produce an output protocol file like this:<br>
 * <table><tr><td>Generation</td><td>Best_Fitness</td><td>Average_Fitness</td><td>Time_elapsed</td></tr>
 * <tr><td>1</td><td>-1340</td><td>-1470</td><td>2</td></tr>
 * <tr><td>13</td><td>-1210</td><td>-1350</td><td>11</td></tr>
 * <tr><td>...</td><td>...</td><td>...</td><td>...</td></tr></table> <br>
 * It could now be useful to extract only the columns <i>Generation</i> and <i>Best_Fitness</i> for
 * further evaluation, e.g. for graphical analysis.
 *****/

public class ExtractColumns {
  private static List tableInformation;

  private static void readInputFile(String name) throws IOException {
    tableInformation = new ArrayList();
    boolean keysRead = false;
    BufferedReader br = new BufferedReader(new FileReader(name));
    String line = br.readLine();
    while (line != null) {
      line = line.trim();
      if ((!line.startsWith("#")) && (line.length() > 0)) {  // ordinary line
        if (!keysRead) {  // keys unread yet, so do this now
          keysRead = true;
          List keys = GlobalMethods.listFromString(line, '\t');
          for (int i = 0; i < keys.size(); i++) {
            String key = (String)keys.get(i);
            tableInformation.add(new KeyValue(key, new ArrayList()));
          }
        }
        else {  // keys have already been read, so proceed with data
          List values = GlobalMethods.listFromString(line, '\t');
          for (int i = 0; i < values.size(); i++) {
            List column = (List)((KeyValue)tableInformation.get(i)).value;
            column.add(values.get(i));
          }
        }
      }
      line = br.readLine();
    }
  }

  private static void removeUnnecessaryInformation(List cols) {
    ListIterator it = tableInformation.listIterator();
    while (it.hasNext()) {
      KeyValue kv = (KeyValue)it.next();
      if (!cols.contains(kv.key))
        it.remove();
    }
    for (int i = 0; i < cols.size(); i++) {
      String key = (String)cols.get(i);
      for (int j = 0; j < tableInformation.size(); j++) {
        KeyValue kv = (KeyValue)tableInformation.get(j);
        if (kv.key.equals(key)) {
          cols.set(i, kv);
          break;
        }
      }
    }
    tableInformation = cols;
  } // end of removeUnnecessaryInformation(List)

  private static void writeOutputFile(String name) throws IOException {
    KeyValue kv;
    FileWriter fw = new FileWriter(name);
    String outString = "";
    int i;
    for (i = 0; i < tableInformation.size(); i++) {
      kv = (KeyValue)tableInformation.get(i);
      outString += kv.key;
      if (i < tableInformation.size() - 1)
        outString += "\t";
      else
        outString += "\n";
    }
    fw.write(outString.toCharArray(), 0, outString.length());
    int index = 0;
    while (true) {
      List list = new ArrayList();
      outString = "";
      for (i = 0; i < tableInformation.size(); i++) {
        kv = (KeyValue)tableInformation.get(i);
        list = (List)kv.value;
        if (index >= list.size())
          break;
        outString += list.get(index).toString();
        if (i < tableInformation.size() - 1)
          outString += "\t";
        else
          outString += "\n";
      }
      if (index >= list.size())
        break;
      fw.write(outString.toCharArray(), 0, outString.length());
      index++;
    }
    fw.close();
  } // end of writeOutputFile(String)

  /**
   * This is the only method of interest. Just run <tt>java tools.ExtractColumns</tt>, and
   * everything will be done for you.
   **/
  public static void main(String args[]) {
    String inFile, outFile, columns;
    try {
      BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
      System.out.print("Input file: ");
      if (args.length >= 1)  // input file specified at command line
        System.out.println(inFile = args[0]);
      else
        inFile = br.readLine();
      System.out.print("Output file: ");
      if (args.length >= 2)  // output file specified at command line
        System.out.println(outFile = args[1]);
      else
        outFile = br.readLine();
      System.out.print("Columns to extract (comma-separated): ");
      if (args.length >= 3)
        System.out.println(columns = args[2]);
      else
        columns = br.readLine();
      List cols = GlobalMethods.listFromString(columns, ',');
      readInputFile(inFile);
      removeUnnecessaryInformation(cols);
      writeOutputFile(outFile);
    } catch (Exception e) {
      System.err.println("An error occurred. Execution stopped.");
      System.out.println(e.toString());
      e.printStackTrace();
      System.exit(0);
    }
  } // end of main(String[])

} // end of class ExtractColumns

