/**
 * Implementation of class "Constraint".
 *
 * Author: Stefan Buettcher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 **/

package ttclasses;

/*****
 * Children of the <tt>Constraint</tt> class implement any kind of constraint
 * that is of interest with respect to the timetabling problem.
 *****/

public abstract class Constraint implements Comparable {
  /**
   * type == <tt>HARD_CONSTRAINT</tt> shows that a TimeTable producing
   * a violation of this very constraint, has to be considered infeasible.
   **/
  public static final int HARD_CONSTRAINT = 1;
  /**
   * type == <tt>SOFT_CONSTRAINT</tt> shows that a TimeTable producing
   * a violation of this very constraint, does not necessarily have to be
   * considered infeasible but rather somehow not so welcome.
   **/
  public static final int SOFT_CONSTRAINT = 2;

  /** The name of the constraint (so that it can be found after creation). **/
  public String name;
  
  /** A reference to the <tt>TTProblem</tt> that this constraint belongs to. **/
  public TTProblem ttProblem;
  
  /** This internal counter will be used by the owning <tt>TTProblem</tt> object
    * in order to keep track of the constraint violations generated by this constraint. **/
  protected int violationCounter;
  
  /** This variable will be used by the owsning <tt>TTProblem</tt> object in order
    * to keep track of the penalty points generated by this constraint. **/
  protected double penaltyPoints;
  
  /** The default constructor sets the problem reference to NULL. **/
  public Constraint() {
    ttProblem = null;
    name = "(No Name)";
  } // end of constructor
  
  /** Lets the caller set the constraint's internal reference to the actual TTProblem. **/
  public void setOwner(TTProblem problem) {
    ttProblem = problem;
  } // end of setOwner(TTProblem)
  
  /**
   * Computes if this constraint is violated by or with respect to the
   * Resource specified by <tt>owner</tt>.
   * @return null, if there is no violation <br>
   *         an instance of <tt>ConstraintViolation</tt> otherwise
   **/
  public abstract ConstraintViolation compute(PlanningClass owner);

  /**
   * Tries to repair the violation specified by <tt>violation</tt>.
   * If no specific violation is indicated, it tries to repair all violations that can
   * be detected.
   * @return true if the violation can be repaired, false otherwise
   **/
  public abstract boolean repair(ConstraintViolation violation);

  /**
   * Returns the type of this constraint (<tt>HARD_CONSTRAINT</tt> or <tt>SOFT_CONSTRAINT</tt>).
   **/
  public abstract int getType();

  /** Implements the functionality of the <tt>compareTo</tt> method defined by <tt>Comparable</tt>. **/
  public int compareTo(java.lang.Object o) {
    return name.compareTo(o.toString());
  }
  
  /** Checks for equality by comparing the names (<tt>String</tt> representations) of two objects. **/
  public boolean equals(java.lang.Object o) {
    return (compareTo(o) == 0);
  }

  /** Simply returns the name. **/
  public String toString() {
    return name;
  }
  
} // end of class Constraint

