/**
 * Implementation of class "ConstraintViolation".
 *
 * Author: Stefan Buettcher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 **/

package ttclasses;

import helperclasses.*;

/*****
 * <tt>ConstraintViolation</tt> stores information about a constraint violation.
 * Instances of this class are created by <tt>Constraint.compute(Resource)</tt>
 * and shall be used for calls to <tt>Constraint.repair(ConstraintViolation)</tt>
 * in order to make repair operations less expensive.
 *****/
public final class ConstraintViolation {
  /** This is the <tt>Constraint</tt> which created the violation. **/
  public Constraint constraint;
  /** This is the <tt>PlanningClass</tt> that the creating constraint belongs to. **/
  public PlanningClass owner;
  /** A textual description of the constraint violation. **/
  public String description;
  /** The penalty points associated with this constraint violation. **/
  public double penalty;
  /** Some additional information that can be used to make reparation work faster. **/
  public Object information;

  /**
   * This constructs a new <tt>ConstraintViolation</tt> instance.
   * @param c   The calling <tt>Constraint</tt> instance.
   * @param own   The owner of the Constraint.
   * @param penalty   The value that defines how much this violation shall be punished.
   **/
  public ConstraintViolation(Constraint c, PlanningClass own, double penalty) {
    constraint = c;
    owner = own;
    description = null;
    this.penalty = penalty;
    information = null;
  }

  /**
   * This constructs a new <tt>ConstraintViolation</tt> instance.
   * @param c   The calling <tt>Constraint</tt> instance.
   * @param own   The owner of the Constraint.
   * @param desc   A short description of the violation. This string is used by the
   *               <tt>toString()</tt> method.
   * @param penalty   The value that defines how much this violation shall be punished.
   **/
  public ConstraintViolation(Constraint c, PlanningClass own, String desc, double penalty) {
    constraint = c;
    owner = own;
    description = desc;
    this.penalty = penalty;
    information = null;
  }

  /** This is just the same as <tt>ConstraintViolation(Contraint, PlanningClass, String, double), except
   *  that we in addition define repair information. **/
  public ConstraintViolation(Constraint c, PlanningClass own, String desc, double penalty, Object information) {
    constraint = c;
    owner = own;
    description = desc;
    this.penalty = penalty;
    this.information = information;
  }

  /** The <tt>repair</tt> method just calls <tt>creator.repair</tt>,
   * where <tt>creator</tt> is the <tt>Constraint</tt> instance
   * that created this object. **/
  public final boolean repair() {
    return constraint.repair(this);
  }

  /** This is equivalent to <tt>constraint.getType()</tt>. **/
  public final int getType() {
    return constraint.getType();
  }

  /**
   * Returns a description of the violation, consisting of a description
   * String originally created by the affiliated Constraint instance and
   * the penalty value.
   **/
  public final String toString() {
    return description + ": " + penalty;
  }

  /** Returns an XML description of the object. **/
  public final String toXML() {
    String result = "<ConstraintViolation name=\"" + GlobalMethods.plainToXML(constraint.toString()) +
                    "\" description=\"" + GlobalMethods.plainToXML(description) + "\" owner=\"" +
                    GlobalMethods.plainToXML(owner.toString()) + "\" type=\"";
    if (getType() == Constraint.HARD_CONSTRAINT)
      result += "Hard Constraints";
    if (getType() == Constraint.SOFT_CONSTRAINT)
      result += "Soft Constraints";
    result += "\" penalty=\"" + penalty + "\"/>\n";
    return result;
  }

} // end of class ConstraintViolation

