/**
 * Implementation of class "Event".
 *
 * Author: Stefan Buettcher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 **/

package ttclasses;

import helperclasses.*;
import java.util.*;

/*****
 * <tt>Event</tt> is the mother of all events that can be managed by the
 * timetable builder.
 *****/

public abstract class Event extends PlanningClass {
  /** Private Random objects will be used for mutation and cross-over. **/
  protected Random random;

  /**
   * This is a list of <tt>ResourceList</tt> instances, representing the different
   * resources types that are used within this very event, e.g. the time slots
   * that are to be used.
   **/
  public ArrayList resources;

  /**
   * This is the number of penalty points created by this event.
   * The value of <tt>penaltyPoints</tt> is updated by the <tt>compute</tt>
   * methods of the constraints involved in the timetabling process.
   **/
  public double penaltyPoints;

  /** An ordinary constructor **/
  public Event() {
    random = GlobalObjects.random;
    resources = new ArrayList();
  }

  /**
   * Returns <tt>true</tt> if there are any time slots used by the Event
   * <tt>e</tt> that are also used by <tt>this</tt> or that are clashing
   * with intervals used by <tt>this</tt>.
   **/
  public final boolean clashes(Event e) {
    ArrayList list1 = getAssignedTimeSlots();
    ArrayList list2 = e.getAssignedTimeSlots();
    for (int j = list1.size() - 1; j >= 0; j--)
      for (int k = list2.size() - 1; k >= 0; k--)
        if (((TimeInterval)list1.get(j)).overlaps((TimeInterval)list2.get(k)))
          return true;
    return false;
  } // end of clashes(Event)

  /**
   * Returns <tt>true</tt> if any of the slots used by this event clashes
   * with the interval specified by <tt>ti</tt>, <tt>false</tt> otherwise.
   **/
  public final boolean clashes(TimeInterval ti) {
    ResourceList list = getResourceList("TimeSlots");
    for (int j = list.size() - 1; j >= 0; j--)
      if (list.isAssigned(j))
        if (ti.overlaps((TimeInterval)list.getResource(j)))
          return true;
    return false;
  } // end of clashes(TimeInterval)

  /**
   * The <tt>updateViews</tt> method updates the event lists of all resources currently
   * in use by calling <tt>resource.addEvent(this)</tt>.
   **/
  protected final void updateViews() {
    for (int i = resources.size() - 1; i >= 0; i--)
      ((ResourceList)resources.get(i)).updateViews();
  }

  /**
   * Events are considered the alleles (atoms) of the chromosome
   * (actual timetable). Thus, they have the ability to mutate randomly.
   **/
  public final void mutate(double mutationRate) {
    for (int i = resources.size() - 1; i >= 0; i--)
      if (random.nextDouble() < mutationRate)
        ((ResourceList)resources.get(i)).mutate();
  }

  /** This method sets the event's genetic code to that of <tt>e</tt>. **/
  public final void setGeneticCode(Event e) {
    ResourceList res1, res2;
    for (int j = resources.size() - 1; j >= 0; j--) {
      res1 = (ResourceList)resources.get(j);
      res2 = (ResourceList)e.resources.get(j);
      res1.setAssignment(res2.getAssignment());
    }
  } // end of setGeneticCode(Event)

  /**
   * Returns the penalty points generated by this event and all resources
   * possibly assignable to the event.
   **/
  public final double getPenaltyPoints() {
    int j, k, size;
    double penalty = 0;
    // first, compute penalty points directly generated by the event
    ArrayList violations = computeConstraintViolations();
    for (j = violations.size() - 1; j >= 0; j--)
      penalty += ((ConstraintViolation)violations.get(j)).penalty;
    // then, sum up those of all resources associated with the event
    for (j = resources.size() - 1; j >= 0; j--) {
      ResourceList resList = (ResourceList)resources.get(j);
      size = resList.size();
      for (k = 0; k < size; k++) 
        penalty += resList.getResource(k).getPenaltyPoints();
    }
    return penalty;
  } // end of getPenaltyPoints()

  /**
   * Performs local search to minimize the penalty points generated within an event.
   **/
  public final void localSearch() {
    for (int j = resources.size() - 1; j >= 0; j--)
      ((ResourceList)resources.get(j)).localSearch();
  } // end of localSearch()

  /** Returns the <tt>ResourceList</tt> with the given name, e.g. "TimeSlots". **/
  public final ResourceList getResourceList(String name) {
    for (int i = 0; i < resources.size(); i++)
      if (((ResourceList)resources.get(i)).getName().equals(name))
        return (ResourceList)resources.get(i);
    return null;
  }

  /** Returns a list of all resources assigned to this event (except the time slots). **/
  public final ArrayList getAssignedResources() {
    ArrayList result = new ArrayList();
    Resource assignment[];
    int cnt, size;
    for (int i = 0; i < resources.size(); i++) {
      ResourceList resList = (ResourceList)resources.get(i);
      if (!resList.getName().equals("TimeSlots")) {
        assignment = resList.getAssignmentArray();
        size = assignment.length;
        for (cnt = 0; cnt < size; cnt++)
          result.add(assignment[cnt]);
      }
    }
    return result;
  } // end of getAssignedResources()

  /** Returns a list of all time slots that have been assigned to this event. **/
  public final ArrayList getAssignedTimeSlots() {
    ResourceList resList = getResourceList("TimeSlots");
    if (resList == null)
      return null;
    Resource list[] = resList.getAssignmentArray();
    ArrayList result = new ArrayList();
    int cnt, size;
    for (int i = list.length - 1; i >= 0; i--) {
      Resource res = list[i];
      if (res instanceof TimeIntervals) {   // encapsulated intervals
        ArrayList l = ((TimeIntervals)res).getTimeIntervals();
        size = l.size();
        for (cnt = 0; cnt < size; cnt++)
          result.add(l.get(cnt));
      }
      else  // pure intervals
        result.add(res);
    }
    return result;
  } // end of getAssignedTimeSlots()

  /** Removes the given resource from the assignments of all resource lists. **/
  public final void freeResource(Resource res) {
    for (int i = resources.size() - 1; i >= 0; i--) {
      ResourceList rl = (ResourceList)resources.get(i);
      rl.freeResource(res);
    }
  } // end of freeResource(Resource)
 
  /** <tt>getSummary()</tt> returns some scheduling information. **/
  public String getSummary() {
    String result = name + ":   ";
    for (int i = resources.size() - 1; i >= 0; i--) {
      ResourceList l = (ResourceList)resources.get(i);
      result += "[" + l.getName() + " (" + l.getMin() + "<" + l.assignedCount() + "<" + l.getMax() + ") = ";
      List allocated = l.getAssignmentList();
      for (int j = allocated.size() - 1; j >= 0; j--) {
        result += allocated.get(j);
        if (j > 0)
          result += " . ";
      }
      result += "] ";
    }
    return result;
  } // end of getSummary()
  
  /** This is part of the <tt>TTProblem.toXML()</tt> feature which translates a timetable to XML. **/
  public String toXML() {
    String result = "<Event name=\"" + GlobalMethods.plainToXML(name) + "\">\n";
    for (int i = 0; i < resources.size(); i++)
      result += ((ResourceList)resources.get(i)).toXML();
    result += "</Event>\n";
    return result;
  } // end of toXML()
    
} // end of class Event


