/**
 * Implementation of class "TTChromosome".
 *
 * Author: Stefan Buettcher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 **/

package ttclasses;

import ga.*;
import helperclasses.*;

import java.io.*;
import java.util.*;
import java.util.zip.*;

/*****
 * The class <tt>TTChromosome</tt> was initially created for testing and comparing
 * purposes only. It was expected to answer the question: Can pure bitstring
 * mutation and cross-over work efficiently in GA-based timetabling algorithms?
 * The answer was: "No." - <tt>TTChromosome</tt> can however be used to encapsulate
 * a TTProblem and thus allow for bigger population sizes (because of the pure
 * bitstring representation).
 *****/
public final class TTChromosome extends GeneticIndividual {
  private TTProblem problem;       // the TTProblem instance we work on
  private boolean data[];          // bitstring representation of the TTProblem instance
  private Random random;           // the usual Random object
  private int geneCount;           // number of Genes
  private int genePositions[];     // necessary information for crossover: at what posititions do the
                                   // distinct genes start/end?

  /** This is an inofficial constructor. Don't use it. **/
  protected TTChromosome(TTProblem ttp, boolean data[], double fitness, int genes[]) {
    problem = ttp;
    setData(data);
    genePositions = genes;
    geneCount = genePositions.length;
    random = GlobalObjects.random;
    this.fitness = fitness;
  }

  /** This is the official constructor. **/
  public TTChromosome(TTProblem ttp) {
    setTTProblem(ttp);
    random = GlobalObjects.random;
    fitness = NO_FITNESS;
  }

  /** Updates the positions of the genes inside the TTProblem instance. **/
  private void updateGeneInformation() {
    int i, j, k, offset, resListCount;
    int eventCount = problem.eventList.size();
    geneCount = 0;
    for (j = 0; j < eventCount; j++)
      geneCount += ((Event)problem.eventList.get(j)).resources.size();
    genePositions = new int[geneCount];
    offset = 0;
    i = 0;
    for (j = 0; j < eventCount; j++) {
      Event e = (Event)problem.eventList.get(j);
      resListCount = e.resources.size();
      for (k = 0; k < resListCount; k++) {
        genePositions[i++] = offset;
        offset += ((ResourceList)e.resources.get(k)).size();
      }
    }
    data = new boolean[offset];
  } // end of updateGeneInformation()

  public void setTTProblem(TTProblem ttp) {
    problem = ttp;
    updateGeneInformation();
    dataToBitstring(ttp);
  }

  public TTProblem getTTProblem() {
    return problem;
  }

  /** Lets the caller set the individual's fitness value. Be cautious! **/
  protected void setFitness(double value) {
    fitness = value;
  }

  /**
   * This method transforms the data of a <tt>TTProblem</tt> instance into the bitstring
   * representation wanted.
   **/
  public void dataToBitstring(TTProblem ttp) {
    int i, j, k, eventCount, resListCount;
    boolean temp[];
    int offset = 0;
    eventCount = ttp.eventList.size();
    for (i = 0; i < eventCount; i++) {
      Event e = (Event)ttp.eventList.get(i);
      resListCount = e.resources.size();
      for (k = 0; k < resListCount; k++) {
        temp = ((ResourceList)e.resources.get(k)).getAssignment();
        for (j = 0; j < temp.length; j++)
          data[offset++] = temp[j];
      }
    }
  } // end of dataToBitstring(TTProblem)

  /* Counterpart of <tt>dataToBitstring(TTProblem)</tt>: Translates the bitstring
   * information back to the TTProblem structures. */
  public TTProblem setDataFromBitstring(TTProblem ttp, boolean data[]) {
    if (ttp.chromosome == this)  // nothing to do: everything is up-to-date
      return ttp;
    int i, j, k, eventCount, resListCount;
    boolean assignment[];
    int offset = 0;
    eventCount = ttp.eventList.size();
    for (i = 0; i < eventCount; i++) {
      Event e = (Event)ttp.eventList.get(i);
      resListCount = e.resources.size();
      for (k = 0; k < resListCount; k++) {
        assignment = ((ResourceList)e.resources.get(k)).getAssignment();
        ((ResourceList)e.resources.get(k)).setAssignment(data, offset);
        offset += assignment.length;
      }
    }
    ttp.fitness = NO_FITNESS;
    ttp.chromosome = this;
    return ttp;
  } // end of setDataFromBitstring(TTProblem, boolean[])


  /** Creates and returns an exact copy of the individual. **/
  public final GeneticIndividual duplicate() {
    return new TTChromosome(problem, data, fitness, genePositions);
  }

  /** Gene-based bitvector crossover. **/
  public final GeneticIndividual crossOver(GeneticIndividual[] partners, int crossingPoints) {
    int i, j, offset, selection, len;
    boolean problems[][] = new boolean[partners.length + 1][];
    boolean select[];
    for (i = 0; i < partners.length; i++)
      problems[i] = ((TTChromosome)partners[i]).getData();
    problems[partners.length] = data;
    int problemCount = problems.length;

    boolean child[] = new boolean[data.length];

    if (crossingPoints == 0) {  // uniform crossover
      for (i = 0; i < geneCount; i++) {
        selection = random.nextInt(problemCount);
        if (i < geneCount - 1)
          len = genePositions[i + 1] - genePositions[i];
        else
          len = data.length - genePositions[i];
        offset = genePositions[i];
        select = problems[selection];
        for (j = 0; j < len; j++) {
          child[offset] = select[offset];
          offset++;
        }
      }
    }
    else if (crossingPoints == 1) {  // one-point crossover
      int point = random.nextInt(geneCount - 1) + 1;
      selection = random.nextInt(partners.length + 1);
      for (i = 0; i < geneCount; i++) {
        if (i == point)
          selection = random.nextInt(partners.length + 1);
        if (i < geneCount - 1)
          len = genePositions[i + 1] - genePositions[i];
        else
          len = data.length - genePositions[i];
        offset = genePositions[i];
        select = problems[selection];
        for (j = 0; j < len; j++) {
          child[offset] = select[offset];
          offset++;
        }
      }
    }
    else if (crossingPoints == 2) {  // two-point crossover
      int point1 = random.nextInt(geneCount - 1) + 1;
      int point2 = random.nextInt(geneCount - 1) + 1;
      selection = random.nextInt(partners.length + 1);
      for (i = 0; i < geneCount; i++) {
        if ((i == point1) || (i == point2))
          selection = random.nextInt(partners.length + 1);
        if (i < geneCount - 1)
          len = genePositions[i + 1] - genePositions[i];
        else
          len = data.length - genePositions[i];
        offset = genePositions[i];
        select = problems[selection];
        for (j = 0; j < len; j++) {
          child[offset] = select[offset];
          offset++;
        }
      }
    } // end of two-point crossover
    else
      throw new UnsupportedOperationException("Two many crossing points.");

    return new TTChromosome(problem, child, NO_FITNESS, genePositions);
  }

  /** Returns the bitstring representation of this chromosome. **/
  protected final boolean[] getData() {
    return data;
  }

  /** Sets the bitstring representation of this chromosome. **/  
  protected final void setData(boolean data[]) {
    if ((this.data == null) || (this.data.length != data.length))
      this.data = new boolean[data.length];
    System.arraycopy(data, 0, this.data, 0, data.length);
    if (problem.chromosome == this)
      problem.chromosome = null;
  }

  public final double getFitness() {
    if (fitness == NO_FITNESS) {
      setDataFromBitstring(problem, data);
      fitness = problem.getFitness();
    }
    return fitness;
  }
  
  public final void mutate(double mutationRate) {
    setDataFromBitstring(problem, data);
    problem.mutate(mutationRate);
    dataToBitstring(problem);
    fitness = problem.getFitness();
  }
  
  public final void performDirectedMutation(double mutationRate) {
    setDataFromBitstring(problem, data);
    problem.performDirectedMutation(mutationRate);
    dataToBitstring(problem);
    fitness = problem.getFitness();
  }    
  
  public final void repair(double repairRate) {
    setDataFromBitstring(problem, data);
    problem.repair(repairRate);
    dataToBitstring(problem);
    fitness = problem.getFitness();
  }

  public final void localSearch() {
    setDataFromBitstring(problem, data);
    problem.localSearch();
    dataToBitstring(problem);
    fitness = problem.getFitness();
  }

  /** Any call is passed to the corresponding <tt>TTProblem</tt> instance. **/
  public final void climb(double climbRate) {
    setDataFromBitstring(problem, data);
    problem.climb(climbRate);
    dataToBitstring(problem);
    fitness = problem.getFitness();
  }

  /**
   * Transforms the assignment bit-vector (<tt>data</tt>) into a byte-vector, according to
   * the definition in <tt>GeneticIndividual</tt>.
   **/
  public final byte[] getGeneticCode() {
    try {
      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      GZIPOutputStream gzos = new GZIPOutputStream(baos);
      ObjectOutputStream oos = new ObjectOutputStream(gzos);
      oos.writeObject(data);
      oos.close();
      return baos.toByteArray();
    } catch (Exception e) {
      System.err.println("Error retrieving Genetic Code: " + e.toString());
      return null;
    }
  } // end of getGeneticCode()

  /** Counterpart to <tt>getGeneticCode()</tt>. **/
  public final void setGeneticCode(byte bytes[]) {
    try {
      ByteArrayInputStream bais = new ByteArrayInputStream(bytes);
      GZIPInputStream gzis = new GZIPInputStream(bais);
      ObjectInputStream ois = new ObjectInputStream(gzis);
      data = (boolean[])ois.readObject();
      fitness = NO_FITNESS;
      if (problem.chromosome == this)
        problem.chromosome = null;
    }
    catch (Exception e) {
      System.err.println("Error setting Genetic Code: " + e.toString());
    }
  } // end of setGeneticCode(byte[])

  public String getSummary() {
    setDataFromBitstring(problem, data);
    return problem.getSummary();
  }

  public String toXML() {
    setDataFromBitstring(problem, data);
    return problem.toXML();
  }

  public String toHTML() {
    setDataFromBitstring(problem, data);
    return problem.toHTML();
  }

  public String getSpeciesDescription() {
    return problem.getSpeciesDescription();
  }

  public final List getProtocolInformation() {
    setDataFromBitstring(problem, data);
    return problem.getProtocolInformation();
  }

  /** For testing purposes only! **/
  public static void main(String args[]) {
    TTChromosome ttc = new TTChromosome(null);
  } // end of main(String[])

} // end of class TTChromosome


