/**
 * Implementation of class "TTProblem".
 *
 * Author: Stefan Buettcher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 **/

package ttclasses;

import ga.*;
import helperclasses.*;

import java.io.*;
import java.text.*;
import java.util.*;

/*****
 * The <tt>TTProblem</tt> class is the super-class of all scheduling
 * problems, e.g. university time-tables.
 *****/
public abstract class TTProblem extends GeneticIndividual implements Cloneable {
  /** A list of all events that have to be scheduled. **/
  protected List eventList;
  /**
   * A list of instances of <tt>ResourceType</tt>. In the School TimeTabling problem,
   * for instance, <tt>resources</tt> contains a list of all teachers,
   * a list of all rooms, and so on.
   **/
  protected List resources;
  /** A list of all constraints that are to be used within a specific problem. **/
  protected List constraints;
  /**
   * The list <tt>globalConstraints</tt> contains references to all constraints
   * that do not belong to a specific <tt>Resource</tt> or, more generally,
   * <tt>PlanningClass</tt> and must therefore be declared <em>global</em>.
   **/
  protected List globalConstraints;
  /**
   * <tt>constraintViolations</tt> is an internal list that stores all computed
   * violations.
   * During the execution of <tt>getFitness()</tt>, all constraints contained in
   * the TTProblem, are asked if a constraint violation occurs. If there is such
   * a violation, it is stored in this list <tt>constraintViolations</tt>.
   **/
  private List constraintViolations;
  /**
   * The timetable that is created by BranchAndBound, genetic evolution or some
   * other TT-solving method, should be able to present itself as a HTML document.
   * Therefore, we need some additional layout information.
   * Example: If we have a weekly school timetable with 5 days a week, 10 timeslots
   * a day, then ttXlayout = 5, ttYlayout = 10, ttZlayout = 1. If we have some other
   * kind of timetable covering a period of 4 weeks with 7 days a week and 3 timeslots
   * per day, then ttXlayout = 7, ttYlayout = 3, ttZlayout = 4.
   **/
  protected int ttXlayout, ttYlayout, ttZlayout;
  /** Just a description of the actual timetabling problem. **/
  protected String description;

  /** This random object is used for mutation and cross-over. **/
  protected Random random;

  /**
   * This field is only used when the TTProblem instance is encapsulated
   * in a TTChromosome and indicates what TTChromosome instance has last
   * accessed the internal data structures of the TTProblem instance.
   **/
  protected TTChromosome chromosome;


  /** An ordinary constructor. **/
  public TTProblem() {
    eventList = new ArrayList();
    resources = new ArrayList();
    constraints = new SortedList();
    globalConstraints = new SortedList();
    constraintViolations = new ArrayList();
    random = GlobalObjects.random;
    ttXlayout = ttYlayout = ttZlayout = 0;
    fitness = NO_FITNESS;
    description = "Unknown Timetabling Problem";
    chromosome = null;
  } // end of our ordinary constructor

  /** Sets the <tt>owner</tt> field of all resources belonging to the timetabling problem to <tt>this</tt>. **/
  protected final void setReferences() {
    int j, k;
    List list;
    for (j = eventList.size() - 1; j >= 0; j--)
      ((Event)eventList.get(j)).setOwner(this);
    for (j = resources.size() - 1; j >= 0; j--) {
      ResourceType rt = (ResourceType)resources.get(j);
      list = rt.components;
      for (k = list.size() - 1; k >= 0; k--)
        ((Resource)list.get(k)).setOwner(this);
    }
    for (j = constraints.size() - 1; j >= 0; j--)
      ((Constraint)constraints.get(j)).setOwner(this);
  } // end of setReferences()

  /** Returns the list of all events that have to be scheduled. **/
  public final List getEvents() {
    return eventList;
  }

  /**
   * Adds a new resource type to the scheduling problem.
   * @param name   The name of the new resource, e.g. "Teacher".
   * @return   A newly created <tt>List</tt> object that will contain
   *           all object of the type specified, e.g. all teachers.
   **/
  public final List addResourceType(String type) {
    ResourceType result = new ResourceType(type);
    resources.add(result);
    return result.components;
  }
  
  /** Returns the resource list that belongs to the name specified. **/
  public final List getResourceType(String type) {
    for (int i = resources.size() - 1; i >= 0; i--)
      if (type.compareTo(((ResourceType)resources.get(i)).name) == 0)
        return ((ResourceType)resources.get(i)).components;
    return null;
  }
  
  /** Returns the interval with index <tt>index</tt>. **/
  public final TimeInterval getTimeSlot(int index) {
    List l = getResourceType("TimeSlot");
    return (TimeInterval)l.get(index);
  }

  /** Returns all resource components of type <tt>type</tt> whose <tt>feature</tt> list contains <tt>feature</tt>. **/
  public final List getResourcesWithFeature(String type, String feature) {
    List list = getResourceType(type);
    if (list == null)
      return null;
    else {
      List l = new ArrayList();
      for (int i = list.size() - 1; i >= 0; i--) {
        Resource r = (Resource)list.get(i);
        if (r.hasFeature(feature))
          l.add(r);
      }
      return l;
    }
  } // end of getResourceWithFeature(String, String)

  /**
   * Returns (if existent) the object with name <tt>name</tt> of the
   * resource type <tt>type</tt>. <br>
   * <i>Example</i> &nbsp; <tt>getComponent("Teacher", "Johnson")</tt> would
   * result in the Resource object "Johnson" within the list of all teachers.
   **/
  public final Resource getComponent(String type, String name) {
    try {
      SortedList l = (SortedList)getResourceType(type);
      return (Resource)l.get(name);
    }
    catch (Exception e) {
      return null;
    }
  } // end of getComponent

  /**
   * Returns (if existent) the object with name <tt>name</tt>. The type of
   * that very resource does not have to be known for this operation.
   **/
  public final Resource getComponent(String name) {
    Object result;
    for (int i = resources.size() - 1; i >= 0; i--) {
      ResourceType rt = (ResourceType)resources.get(i);
      result = rt.get(name);
      if (result != null)
        return (Resource)result;
    }
    return null;
  } // end of getComponent(String)

  /** Returns the list of the resource lists, i.e. the list of all time slots, the list of all rooms, ... **/
  public final List getResources() {
    return resources;
  }

  /** Returns a reference to the constraint with the given name. **/
  public final Constraint getConstraint(String name) {
    Constraint result = (Constraint)((SortedList)constraints).get(name);
    if (result == null)
      System.err.println("Unable to find Constraint \"" + name + "\".");
    return result;
  }
  
  /**
   * Since all events and resources can be created dynamically while processing
   * the XML data file belonging to the GTL description of the problem, <tt>TTProblem</tt>
   * must be able to dynamically instantiate objects of certain types.
   **/
  public Resource createResource(String type, String name) {
    if (type.equals("TimeSlot"))
      return new TimeInterval(name);
    else if (type.equals("TimeSlots"))
      return new TimeIntervals(name);
    else
      return null;
  } // end of createResource(String, String)

  /**
   * Since all events and resources can be created dynamically while processing
   * the XML data file belonging to the GTL description of the problem, <tt>TTProblem</tt>
   * must be able to dynamically instantiate objects of certain types.
   **/
  public abstract Event createEvent(String type, String name);

  /**
   * A timetabling-problem definition (using the General Timetabling Language) consists
   * of two parts: the structural declaration (type declaration, ...) and the instantiation,
   * describing the actual problem setup. <br>
   * <tt>initialize</tt> loads these instantiation data from the given input file.
   * @param inputFile   name of the file containing the necessary information
   **/
  public final void initialize(String inputFile) {
    XMLParser.XMLData xmlData = new XMLParser().parse(inputFile);
    if (xmlData == null)
      System.err.println("Error parsing XML file.");
    else {
      processXMLData(xmlData);
      setReferences();
    }
  }

  /** Just another possibility to create all objects involved in the timetabling process. **/
  public final void initialize(char inputData[]) {
    XMLParser.XMLData xmlData = new XMLParser().parse(inputData);
    processXMLData(xmlData);

    // set correct reference to TTProblem in all constraints
    for (int i = 0; i < constraints.size(); i++)
      ((Constraint)constraints.get(i)).ttProblem = this;
  }

  /** This is the actual processing of the XML data. **/
  private final void processXMLData(XMLParser.XMLData data) {
    int errorNumber = -1;

    System.out.print("[Processing XML data file, ");

    try {
    if (data == null) {
      System.err.println("Could not read XML data for TTProblem. Unable to proceed.");
      return;
    }
    if (data.item.name.equals("TTProblem")) {  // everything alright so far...
      XMLParser.XMLItem item;
      int j, k, size;
      List l = data.item.subItems;
      size = l.size();

      // set layout (X, Y, Z)
      List layout = GlobalMethods.listFromXMLString(data.item.getAttributeValue("layout"));
      if(layout.size() !=3)
        System.err.println("Error: Layout dimension must have 3 components");
      ttXlayout = Integer.parseInt((String)layout.get(0));
      ttYlayout = Integer.parseInt((String)layout.get(1));
      ttZlayout = Integer.parseInt((String)layout.get(2));

      globalConstraints = new SortedList();
      String globals = data.item.getAttributeValue("constraints");
      if ((globals == null) || (globals.length() == 0))
        globals = data.item.getAttributeValue("globalConstraints");
      if ((globals != null) && (globals.length() > 0)) {
        List globalsListed = GlobalMethods.listFromXMLString(globals);
        for (int cnt = 0; cnt < globalsListed.size(); cnt++)
          globalConstraints.add(getConstraint((String)globalsListed.get(cnt)));
      }

      errorNumber = 0;

      description = (String)data.item.getAttributeValue("description");
      if (description == null)
        description = "Unknown Timetabling Problem";

      // process constraint data
      for (j = 0; j < size; j++) {
        item = (XMLParser.XMLItem)l.get(j);
        if (item.name.equals("Constraint"))
          globalConstraints.add(getConstraint(item.name));
      }

      ((SortedList)globalConstraints).sortList();

      errorNumber = 1;

      // process resource data
      for (j = 0; j < size; j++) {
        item = (XMLParser.XMLItem)l.get(j);

        if (item.name.equals("Resource")) {
          // System.out.print("Processing Resource " + item.getAttributeValue("name") + ", ");
          String type = GlobalMethods.XMLtoPlain(item.getAttributeValue("type"));
          String name = GlobalMethods.XMLtoPlain(item.getAttributeValue("name"));

          Resource res = createResource(type, name);
          res.owner = this;

          // Resource with that name not found
          if(res == null)
              System.err.println("Resource with name " + name + " not specified!");

          for (k = 0; k < item.attributes.size(); k++) {
            XMLParser.XMLAttribute attr = (XMLParser.XMLAttribute)item.attributes.get(k);
            // System.out.print("Attributes: " + attr.name + ", Value: " + attr.value);

            // type attribute only for resource identification
            if(!attr.name.equals("type"))
              res.setVariable(attr.name, attr.value);
          }
          // System.out.println();
        }
      } // end of outer loop

      errorNumber = 2;

      // process event data
      for (j = 0; j < size; j++) {
        item = (XMLParser.XMLItem)l.get(j);
        if (item.name.equals("Event")) {
          // System.out.print("Processing Event " + item.getAttributeValue("name") + ", Resourcelists:");
          String type = GlobalMethods.XMLtoPlain(item.getAttributeValue("type"));
          String name = GlobalMethods.XMLtoPlain(item.getAttributeValue("name"));
          Event ev = createEvent(type, name);
          ev.owner = this;
          for (k = 0; k < item.attributes.size(); k++) {
            XMLParser.XMLAttribute attr = (XMLParser.XMLAttribute)item.attributes.get(k);
            if(!attr.name.equals("type"))
              ev.setVariable(attr.name, attr.value);
          }

          // after the initialization of event-wide values, loop through
          // all ResourceList objects
          for (k = 0; k < item.subItems.size(); k++) {
            XMLParser.XMLItem subItem = (XMLParser.XMLItem)item.subItems.get(k);
            if (subItem.name.equals("ResourceList")) {
              // process attribute values
              // System.out.print(" " + subItem.getAttributeValue("name"));
              String min = GlobalMethods.XMLtoPlain(subItem.getAttributeValue("min"));
              String max = GlobalMethods.XMLtoPlain(subItem.getAttributeValue("max"));
              String fixed = GlobalMethods.XMLtoPlain(subItem.getAttributeValue("fixed"));
              // and translate the String values in the takeFrom list to Resource references
              List list1 = GlobalMethods.listFromXMLString(subItem.getAttributeValue("takeFrom"));
              List list2 = new ArrayList();
              for (int m = 0; m < list1.size(); m++) {
                Resource r = getComponent((String)list1.get(m));
                if (r != null)
                  list2.add(r);
                else {
                  System.err.println("Unknown resource " + (String)list1.get(m) + " in event " + name + "!");
new Exception().printStackTrace();
                }
              }
              // create ResourceList
              ResourceList rl = new ResourceList(ev, GlobalMethods.XMLtoPlain(subItem.getAttributeValue("name")),
                                Integer.parseInt(min), Integer.parseInt(max), list2);
              // in case, fix is required, do so
              if ((fixed != null) && (fixed.toLowerCase().equals("true")))
                rl.fix();
              // and eventually add to the ResourceList list
              ev.resources.add(rl);
            }
          }
          // System.out.println();
        }
      } // end of outer loop

      System.out.println("done.]");

    }
    else {  // first-level XML item is not of type "TTProblem" => Error!
      System.out.println("error.]");
      System.err.println("Wrong file format when processing XML data for TTProblem. Unable to proceed.");
      return;
    }

  } catch (Exception e) {
    System.err.println("/--------------------------------------------------------\\");
    System.err.println("| Error while parsing XML input. This is probably due to |");
    System.err.println("| an incorrect XML syntax. Unable to proceed.            |");
    System.err.println("\\--------------------------------------------------------/");
    switch(errorNumber) {
      case -1:
        System.err.println("Error setting layout");
        break;
      case 0:
        System.err.println("Error processing constraint data");
        break;
      case 1:
        System.err.println("Error processing resource data");
        break;
      case 2:
        System.err.println("Error processing event data");
        break;
    }
    e.printStackTrace(System.err);
    System.exit(0);
  } } // end of processXMLData(XMLData)

  /**
   * Calls the <tt>compute</tt> method of all constraints either belonging to a specific
   * <tt>PlanningClass</tt> instance or being references from within the list
   * <tt>globalConstraints</tt>. All resulting <tt>ConstraintViolation</tt> instances
   * will be stored in <tt>constraintViolations</tt>.
   * @return the sum of all penalty points created by the constraints
   **/
  public final double computeConstraintViolations() {
    int j, k;
    Resource res;
    Event ev;
    constraintViolations = new ArrayList();

    // loop through all planning classes and compute the constraint violations
    for (j = eventList.size() - 1; j >= 0; j--) {
      ev = (Event)eventList.get(j);
      ev.penaltyPoints = 0;
      try {
        constraintViolations.addAll(ev.computeConstraintViolations());
      }
      catch (Exception e) {
        System.err.println("Exception in computing constraint violations with Event " + ev.name);
        System.err.println(e.toString());
      }
    }

    for (j = resources.size() - 1; j >= 0; j--) {
      List l = ((ResourceType)resources.get(j)).components;
      for (k = l.size() - 1; k >= 0; k--) {
        res = (Resource)l.get(k);
        try {
          constraintViolations.addAll(res.computeConstraintViolations());
        }
        catch (Exception e) {
          System.err.println("Exception in computing constraint violations with Resource " + res.name);
          System.err.println(e.toString());
        }
      }
    }

    // loop through all global constraints
    for (j = globalConstraints.size() - 1; j >= 0; j--)
      constraintViolations.add(((Constraint)globalConstraints.get(j)).compute(null));
    
    // sum up all penalty points
    double result = 0;
    for (j = constraintViolations.size() - 1; j >= 0; j--) {
      ConstraintViolation cv = (ConstraintViolation)constraintViolations.get(j);
      Constraint c = cv.constraint;
      result += cv.penalty;
    }

    return result;
  } // end of computeConstraintViolations
  
  /**
   * Returns the fitness of the individual (TimeTable) by calculating all constraint violations and
   * adding their penalty values. A list of all violations will be kept stored in <tt>constraintViolations</tt>.
   **/
  public final double getFitness() {
    // if the current fitness is unknown, re-calculate it now
    if (fitness == NO_FITNESS)
      fitness = -computeConstraintViolations();
    return fitness;
  } // end of getFitness(int)

  /**
   * <i>Defined in class GeneticIndividual</i> <br>
   * The repair method allows to use problem-specific knowledge in order
   * to (hopefully) improve the current solution.
   **/
  public final void repair(double repairRate) {
    if (fitness == NO_FITNESS)
      getFitness();
    if (constraintViolations != null)
      for (int i = constraintViolations.size() - 1; i >= 0; i--)
        if (random.nextDouble() < repairRate)
          ((ConstraintViolation)constraintViolations.get(i)).repair();
    fitness = NO_FITNESS;
  }

  /**
   * <i>Defined in class GeneticIndividual</i> <br>
   * Optimization through local search. The idea is that hopefully the events
   * are not linked too closely so that each event is optimizable individually.
   * Local search is quasi-exhaustive and should therefore not be used too
   * frequently during the evolution process. It is suggested to use it after
   * the actual evolution instead.
   * @see GeneticIndividual
   **/
  public final void localSearch() {
    System.out.print("[Starting local search");
    int size = eventList.size();
    int newPercentage, percentage;
    int oldPercentage = 0;
    for (int j = 0; j < size; j++) {
      newPercentage = j * 100 / size;
      percentage = oldPercentage;
      while (newPercentage >= percentage + 10)
        percentage += 10;
      if (percentage != oldPercentage) {
        oldPercentage = percentage;
        System.out.print(".." + percentage + "%");
      }
      ((Event)eventList.get(j)).localSearch();
    }
    System.out.println("..100%]");
    fitness = NO_FITNESS;
  } // end of localSearch()

  /**
   * The method <tt>climb</tt> performs some sort of hill-climbing by
   * applying local search mechanisms to the current timetable.
   **/
  public final void climb(double climbRate) { 
    int selection = random.nextInt(eventList.size());
    ((Event)eventList.get(selection)).localSearch();
    fitness = NO_FITNESS;
  } // end of climb()

  /** Creates an exact copy of the individual. **/
  public final GeneticIndividual duplicate() {
    TTProblem result = (TTProblem)createNewInstance();
    List events = result.getEvents();
    for (int i = eventList.size() - 1; i >= 0; i--) {
      Event e1 = (Event)eventList.get(i);
      Event e2 = (Event)events.get(i);
      e2.setGeneticCode(e1);
    }
    return result;
  } // end of duplicate()

  /**
   * <tt>updateViews()</tt> sets the respective views (of the resources) correctly
   * so that there is no inconsistency between events and resources.
   **/
  public final void updateViews() {
    int j, k;
    List l;
    for (j = 0; j < resources.size(); j++) {
      l = ((ResourceType)resources.get(j)).components;
      for (k = l.size() - 1; k >= 0; k--)
        ((Resource)l.get(k)).clearAssignments();
    }
    for (j = eventList.size() - 1; j >= 0; j--)
      ((Event)eventList.get(j)).updateViews();
    fitness = NO_FITNESS;
  } // end of updateViews()

  /** Produces a new timetable using the crossover operator. **/
  public final GeneticIndividual crossOver(GeneticIndividual[] partners, int crossingPoints) {
    int i, j;
    TTProblem selectedIndividual;
    TTProblem result = (TTProblem)duplicate();
    if (crossingPoints == 0) {  // uniform cross-over
      // granularity is RESOURCE_LIST, i.e. every resource list is considered an
      // independent gene / allele
      for (i = eventList.size() - 1; i >= 0; i--) {
        // loop through all events and all resource list; compose the new individual
        // of randomly-chosen resource lists
        Event resultEvent = (Event)result.eventList.get(i);
        for (j = resultEvent.resources.size() - 1; j >= 0; j--) {
          int selection = random.nextInt(partners.length + 1);
          if (selection < partners.length)
            selectedIndividual = (TTProblem)partners[selection];
          else
            selectedIndividual = this;
          ResourceList targetResList = (ResourceList)resultEvent.resources.get(j);
          ResourceList sourceResList = (ResourceList)((Event)selectedIndividual.eventList.get(i)).resources.get(j);
          targetResList.setAssignment(sourceResList.getAssignment());
        }
      } // end of loop through events
    }
    else if ((crossingPoints == 1) || (crossingPoints == 2)) {
      // one-point or two-point crossover
      // granularity is EVENT, i.e. there will be no event as part of the newly-created
      // individual that is composed of resource lists from different parent individuals
      int point1, point2;
      point1 = random.nextInt(eventList.size() + 1);
      if (crossingPoints == 2)
        point2 = random.nextInt(eventList.size() + 1);
      else
        point2 = -1;
      int selection = random.nextInt(partners.length + 1);
      if (selection < partners.length)
        selectedIndividual = (TTProblem)partners[selection];
      else
        selectedIndividual = this;
      for (i = eventList.size() - 1; i >= 0; i--) {
        if ((i == point1) || (i == point2)) {
          selection = (selection + 1) % (partners.length + 1);
          if (selection < partners.length)
            selectedIndividual = (TTProblem)partners[selection];
          else
            selectedIndividual = this;
        }
        Event geneticCode = ((Event)selectedIndividual.eventList.get(i));
        ((Event)result.eventList.get(i)).setGeneticCode(geneticCode);
      }
    }
    else  // more than 2 crossover points are not supported
      throw new UnsupportedOperationException();
    return result;
  } // end of crossOver(GeneticIndividual[], int)

  /** Lets the individual (schedule) mutate with the given mutation rate. **/
  public final void mutate(double mutationRate) {
    for (int i = eventList.size() - 1; i >= 0; i--)
      ((Event)eventList.get(i)).mutate(mutationRate);
    fitness = NO_FITNESS;
  }

  /**
   * This method oranizes some kind of <em>directed mutation</em>. Certain events
   * which are making trouble are mutated in a direct way. The mutation rate for every
   * single event is directly proportional to the event's penalty points.
   **/
  public final void performDirectedMutation(double mutationRate) {
    int i; // a counter
    int k = eventList.size(); // for performance purposes
    double average = 0; // the average penalty points
    double min = Double.POSITIVE_INFINITY;
    double max = Double.NEGATIVE_INFINITY;
    Event worstEvent = null;
    if (fitness == NO_FITNESS)
      getFitness();
    for (i = 0; i < k; i++) {
      // collect all events' penalty points in order to calculate min, max (for scaling)
      // and average penalty points
      Event e = (Event)eventList.get(i);
      average += e.penaltyPoints;
      if (e.penaltyPoints < min)
        min = e.penaltyPoints;
      if (e.penaltyPoints > max) {
        worstEvent = e;
        max = e.penaltyPoints;
      }
    }
    average /= k;
    average -= min;
    for (i = 0; i < k; i++) {
      Event e = (Event)eventList.get(i);
      double effectivePenalty = e.penaltyPoints - min;
      if (random.nextDouble() < mutationRate * effectivePenalty / average)
        e.mutate(mutationRate * effectivePenalty / average);
      if (e == worstEvent)
        e.mutate(0.5);
    }
    fitness = NO_FITNESS;
  } // end of performDirectedMutation(double)

  /**
   * <tt>createNewInstance()</tt> creates a new object of the scheduling
   * problem that is to be solved. This is, for instance, necessary when
   * running a Genetic Algorithm to solve the problem because we need
   * to generate an initial population.
   **/
  public abstract TTProblem createNewInstance();

  /**
   * Solves the Timetabling problem by calling the solver's solve method.
   **/
  public final TTProblem solve(TTSolver solver) {
    solver.setProblem(this);
    return solver.solve();
  }

  /**
   * Returns a description of the current timetable solution.
   **/
  public final String getSummary() {
    if (fitness == NO_FITNESS)
      getFitness();
    String result = "\n[Final Timetable: " + fitness + "]\n";

    String dummy = "";

    for (int i = eventList.size() - 1; i >= 0; i--) {
      dummy += ((Event)eventList.get(i)).getSummary() + "\n";
      if (dummy.length() > 10000) {
        result += dummy;
        dummy = "";
      }
    }
    dummy += "\n\n[Constraint Violations (Hard Constraints)]\n";
    for (int i = 0; i < constraintViolations.size(); i++)
      if (((ConstraintViolation)constraintViolations.get(i)).getType() == Constraint.HARD_CONSTRAINT) {
        dummy += ((ConstraintViolation)constraintViolations.get(i)).toString() + "\n";
        if (dummy.length() > 10000) {
          result += dummy;
          dummy = "";
        }
      }
    dummy += "\n\n[Constraint Violations (Soft Constraints)]\n";
    for (int i = 0; i < constraintViolations.size(); i++)
      if (((ConstraintViolation)constraintViolations.get(i)).getType() == Constraint.SOFT_CONSTRAINT) {
        dummy += ((ConstraintViolation)constraintViolations.get(i)).toString() + "\n";
        if (dummy.length() > 10000) {
          result += dummy;
          dummy = "";
        }
      }
    return result + dummy;
  } // end of getSummary()

  /**
   * Returns an XML description of the timetable [inherited from GeneticIndividual].
   **/
  public final String toXML() {
    if (fitness == NO_FITNESS)
      getFitness();
    int j, k, l;
    String result = "";  // this is where the final result will be stored
    String dummy = "";  // we need this for improving the performance
    l = eventList.size();
    for (j = 0; j < l; j++) {
    // first, get the XML code for all events
      dummy += ((Event)eventList.get(j)).toXML();
      if (dummy.length() > 10000) {
      // as soon as dummy contains more than 10000 chars, flush
        result += dummy;
        dummy = "";
      }
    }
    for (j = 0; j < resources.size(); j++) {
    // now, get the XML code of all resources
      String type = ((ResourceType)resources.get(j)).name;
      List list = ((ResourceType)resources.get(j)).components;
      l = list.size();
      for (k = 0; k < l; k++) {
        dummy += ((Resource)list.get(k)).toXML(type);
        if (dummy.length() > 10000) {
          result += dummy;
          dummy = "";
        }
      }
    }
    l = constraintViolations.size();
    for (j = 0; j < l; j++) {
    // last, get the XML code of all constraint violations
      dummy += ((ConstraintViolation)constraintViolations.get(j)).toXML();
      if (dummy.length() > 10000) {
        result += dummy;
        dummy = "";
      }
    }
    return "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n" +
           "<TimeTable layout=\"" + ttXlayout + "," + ttYlayout + "," + ttZlayout +
           "\" description=\"" + GlobalMethods.plainToXML(description) + "\">\n" +
           result + dummy + "\n</TimeTable>\n";
  } // end of toXML()

  
  /** Generates the HTML code of an individual timetable for the given resource. **/
  public final String createIndividualHTML(Resource res) {
    String slotText[][][] = new String[ttXlayout + 1][ttYlayout + 1][ttZlayout + 1];
    int j, k, l, m;

    String result = "";

    List slotList = getResourceType("TimeSlot");
    TimeInterval slots[][][] = new TimeInterval[ttXlayout+1][ttYlayout+1][ttZlayout+1];
    try {
      for (j = slotList.size() - 1; j >= 0; j--) {
        TimeInterval ti = (TimeInterval)slotList.get(j);
        slots[ti.layoutX][ti.layoutY][ti.layoutZ] = ti;
      }
    } catch (Exception e) {
      result += "[There was an error. Could not create HTML output.]\n";
      return result;
    }

    for (k = 1; k <= ttXlayout; k++)
      for (l = 1; l <= ttYlayout; l++)   // first, reset the text to be displayed for each timeslot
        for (m = 1; m <= ttZlayout; m++)
          slotText[k][l][m] = "<p align=\"right\"><b>" + slots[k][l][m].name + "</b></p>";

    List events = res.getEvents();
    for (int evCnt = 0; evCnt < events.size(); evCnt++) {
      Event ev = (Event)events.get(evCnt);   // then, loop through all events that this resource has been assigned to
      String name = ev.name;
      List timeslots = ev.getAssignedTimeSlots();
      for (int rlCnt = 0; rlCnt < timeslots.size(); rlCnt++) {  // determine the timeslots occupied by these events
        TimeInterval ti = (TimeInterval)timeslots.get(rlCnt);
        slotText[ti.layoutX][ti.layoutY][ti.layoutZ] += "<a href=\"#" + name.toLowerCase().replace(' ', '_') +
                                                        "\">" + name + "</a><br>";   // and update the accordings texts
      }
    }

    // output the just-built timetable for the resource
    result += "<u>Individual Timetable for " + res.toString() + "</u> <br><br>\n";

    if (events.size() == 0)
      result += "(no events scheduled) <br><br>\n";
    else
      for (k = 1; k <= ttZlayout; k++) {
        result += "<table border=\"3\" cellpadding=\"2\" cellspacing=\"2\">\n";
        for (l = 1; l <= ttYlayout; l++) {
          result += "  <tr>\n";
          for (m = 1; m <= ttXlayout; m++)
            result += "    <td valign=\"top\">" + slotText[m][l][k] + "</td>\n";
          result += "  </tr>\n";
        }
        result += "</table> <br><br>\n";
      } // end of Z-loop for individual view

    result += "<u>Constraint Violations</u> <br>\n";

    List violations = res.computeConstraintViolations();
    for (k = 0; k < violations.size(); k++) {
      ConstraintViolation cv = (ConstraintViolation)violations.get(k);
      result += cv.description + " (" + cv.penalty + ")<br>\n";
    }

    result += "<br> <hr> <br><br>\n";
    return result;
  } // end of createIndividualHTML(Resource)

  /** Generates a beatiful HTML output of the timetable just created. **/
  public final String toHTML() {
    String result = "<html><head><title>Timetable</title>\n";
    result += "<body bgcolor=\"#FFFFFF\" text=\"#000000\">\n";

    // x, y, z > 0 means: a layout has been defined and will be used for output

    if ((ttXlayout > 0) || (ttYlayout > 0) || (ttZlayout > 0)) {  // let's create some HTML output
      int j, k, l, m;
      List slotList = getResourceType("TimeSlot");
      TimeInterval slots[][][] = new TimeInterval[ttXlayout+1][ttYlayout+1][ttZlayout+1];
      try {
        for (j = slotList.size() - 1; j >= 0; j--) {
          TimeInterval ti = (TimeInterval)slotList.get(j);
          slots[ti.layoutX][ti.layoutY][ti.layoutZ] = ti;
        }
      } catch (Exception e) {
        result += "[There was an error. Could not create HTML output.]\n";
      }

      result += "<b>General View: All events</b> <br><br>\n";
      for (j = 1; j <= ttZlayout; j++) {
        result += "<table border=\"3\" cellpadding=\"2\" cellspacing=\"2\">\n";
        for (k = 1; k <= ttYlayout; k++) {
          result += "  <tr>\n";
          for (l = 1; l <= ttXlayout; l++) {
            TimeInterval ti = slots[l][k][j];
            if (ti == null)
              result += "    <td valign=\"top\"><p align=right><b>(no TimeSlot)</b></p>";
            else {
              result += "    <td valign=\"top\"><p align=right><b>" + ti.name + "</b></p>";
              List events = ti.getEvents();
              for (m = events.size() - 1; m >= 0; m--) {
                String name = ((Event)events.get(m)).name;
                result += "<a href=\"#" + name.toLowerCase().replace(' ', '_') + "\">" + name + "</a><br>";
              }
            }
            result += "</td>\n";
          }
          result += "  </tr>\n";
        }
        result += "</table> <br><br>\n";
      } // end of Z-loop for general view

      result += "<hr> <br><br>\n";

      // now, output the timetables (views) of every component involved in the process

      String individualView = "";  // to contain the view of every individual resource

      for (j = 0; j < resources.size(); j++) {
        if (!((ResourceType)resources.get(j)).name.equals("TimeSlot")) {
          List list = ((ResourceType)resources.get(j)).components;
          for (int cnt = 0; cnt < list.size(); cnt++) {
            // compute individual timetable
            individualView += createIndividualHTML((Resource)list.get(cnt));
            if (individualView.length() > 12000) {
              result += individualView;
              individualView = "";
            }
          }  // end of loop through all resources of one type
        }
      } // end of loop through all resources lists (types of resources)

      if (individualView.length() > 0) {
        result += individualView;
        individualView = "";
      }

      // finally, output all event assignments
      l = eventList.size();
      for (k = 0; k < l; k++) {
        Event ev = (Event)eventList.get(k);

        individualView += "<u>Resource Assignment for " + ev.name + "</u> <br><br>\n";

        for (m = 0; m < ev.resources.size(); m++) {
          ResourceList rl = (ResourceList)ev.resources.get(m);
          individualView += rl.getName() + ": ";
          individualView += rl.getAssignmentList().toString() + "\n";
        }

        individualView += "<br><hr> <br><br>\n";

        if (individualView.length() > 12000) {
          result += individualView;
          individualView = "";
        }
      }

      if (individualView.length() > 0) {
        result += individualView;
        individualView = "";
      }

    }
    else
      result += "[No layout specified => No HTML output.]\n";

    return result + "</body>\n</html>\n";
  } // end of toHTML()

  /** Return some information about the number of events, etc. **/
  public final String getSpeciesDescription() {
    int akku = 0;
    for (int i = 0; i < resources.size(); i++) {
      List l = ((ResourceType)resources.get(i)).components;
      akku += l.size();
    }
    return "Species: TTProblem. Events: " + eventList.size() + ", Resources: " + akku;
  } // end of getSpeciesDescription

  /** Returns some information on constraint violation distribution. **/
  public final List getProtocolInformation() {
    ArrayList result = new ArrayList();
    int i, size;
    size = constraints.size();
    for (i = 0; i < size; i++) {
      // first, set violation count to zero for all constraints
      Constraint c = (Constraint)constraints.get(i);
      c.violationCounter = 0;
      c.penaltyPoints = 0;
    }
    computeConstraintViolations();
    size = constraintViolations.size();
    for (i = 0; i < size; i++) {
      // then, update the violation count according to the constriant violations encountered
      ConstraintViolation cv = (ConstraintViolation)constraintViolations.get(i);
      cv.constraint.violationCounter++;
      cv.constraint.penaltyPoints += cv.penalty;
    }
    size = constraints.size();
    DecimalFormat decFormat = new DecimalFormat("0.000");
    for (i = 0; i < size; i++) {
      // finally, store the numbers in the result list
      Constraint c = (Constraint)constraints.get(i);
      result.add(new KeyValue(c.name + "_Count", new Integer(c.violationCounter)));
      result.add(new KeyValue(c.name + "_Penalty", decFormat.format(c.penaltyPoints)));
    }
    return result;
  } // end of getProtocolInformation()

  /** <tt>saveToDisk</tt> writes the <tt>solution</tt> String to <tt>filename</tt>. **/
  public static void saveToDisk(String solution, String fileName) {
    try {
      OutputStream os = new FileOutputStream(fileName);
      os.write(solution.getBytes());
      os.flush();
    } catch (IOException ioe) {
      System.err.println("\nAn IOException writing file " + fileName + ".\n");
      System.err.println(ioe.toString());
    }
  } // end of saveToDisk(String, String)

} // end of class TTProblem

