/**
 * Implementation of class "TimeIntervals".
 *
 * Author: Stefan Buettcher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 **/

package ttclasses;

import helperclasses.*;
import java.util.*;

/*****
 * This class manages encapsulation of <tt>TimeInterval</tt> objects. This might be
 * necessary (or at least advantageous) when individual time intervals are to be bundled.
 * Example: In the Highschool Timetabling Problem, we have lessons that are combined
 * to two-hour units ("Doppelstunden"). Timeslot assignment must respect that those units
 * must not be interrupted by long breaks, and so on.
 *****/
public final class TimeIntervals extends Resource implements Comparable {
  private ArrayList timeSlots;  // the different time intervals managed by the object
  private int size, duration;   // number of time slots and the duration at all (summed up)
  private boolean updated;      // string (name) -> TimeInterval reference conversion already done?

  public TimeIntervals() {
    name = "(no name)";
    timeSlots = new ArrayList();
    size = 0; duration = 0;
    owner = null;
    updated = false;
  }

  public TimeIntervals(String name) {
    this.name = name;
    timeSlots = new ArrayList();
    size = 0; duration = 0;
    owner = null;
    updated = false;
  }

  /** Returns the list of all time slots contained. **/
  public ArrayList getTimeIntervals() {
    return timeSlots;
  }

  /** Sets the references to the TimeInterval objects specified by their names. **/
  private void updateStringsToTimeIntervals() {
    if (owner == null) return;
    Object o;
    TimeInterval ti;
    duration = 0;
    for (int i = 0; i < size; i++) {
      o = timeSlots.get(i);
      if (o instanceof String) {
        String s = (String)o;
        ti = (TimeInterval)owner.getComponent("TimeSlot", s);
        timeSlots.set(i, ti);
      }
      else
        ti = null;
      if (ti != null)
        duration += ti.getDuration();
    }
    updated = true;
  } // end of updateStringsToTimeIntervals()

  public void setVariable(String key, String value) {
    if (key.equals("timeSlots")) {
      timeSlots = GlobalMethods.listFromXMLString(value);
      size = timeSlots.size(); duration = 0;
      updated = false;
    }
    else
      super.setVariable(key, value);
  }

  public void addTimeSlot(TimeInterval ti) {
    timeSlots.add(ti);
    size++;
    duration += ti.getDuration();
  }

  /**
   * The <tt>overlaps</tt> method checks if one of the intervals encapsulated in this
   * list is overlapped by the one specified by <tt>ti</tt>.
   * @return true if intervals overlap, false otherwise
   **/
  public final boolean overlaps(TimeInterval ti) {
    if (!updated)
      updateStringsToTimeIntervals();
    TimeInterval slot;
    for (int i = 0; i < size; i++) {
      slot = (TimeInterval)timeSlots.get(i);
      if (slot.overlaps(ti))
        return true;
    }
    return false;
  }

  /**
   * The <tt>overlaps</tt> method checks if one of the intervals encapsulated in <tt>this</tt>
   * list is overlapped by one of the intervals encapsulated in <tt>ts</tt>.
   * @return true if intervals overlap, false otherwise
   **/
  public final boolean overlaps(TimeIntervals ti) {
    if (!updated)
      updateStringsToTimeIntervals();
    TimeInterval slot;
    for (int i = 0; i < size; i++) {
      slot = (TimeInterval)timeSlots.get(i);
      if (ti.overlaps(slot))
        return true;
    }
    return false;
  }

  /** Returns the duration of the time intervals encapsulated by this list in minutes. **/
  public final int getDuration() {
    if (!updated)
      updateStringsToTimeIntervals();
    return duration;
  }

  protected final void clearAssignments() {
    if (!updated)
      updateStringsToTimeIntervals();
    for (int i = 0; i < size; i++)
      ((TimeInterval)timeSlots.get(i)).clearAssignments();
    super.clearAssignments();
  }

  protected final void addAssignment(Assignment ass) {
    if (!updated)
      updateStringsToTimeIntervals();
    for (int i = 0; i < size; i++)
      ((TimeInterval)timeSlots.get(i)).addAssignment(ass);
    super.addAssignment(ass);
  }

  protected final void removeAssignment(Assignment ass) {
    if (!updated)
      updateStringsToTimeIntervals();
    for (int i = 0; i < size; i++)
      ((TimeInterval)timeSlots.get(i)).removeAssignment(ass);
    super.removeAssignment(ass);
  }

} // end of class TimeIntervals


